#!/usr/bin/env python
"""A class for the graphical interface used to review the predicted conformers."""

# Copyright 2010, 2011, 2012 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

import gtk
import pango
from copy import deepcopy
#from pprint import pprint     #for debugging
#from cPickle import dump      #for debugging
import os
#import sys
from time import sleep      #for debugging

from coot import set_rotation_centre, add_atom_label, remove_atom_label, graphics_draw, close_generic_object, set_display_generic_object

from puckerList import puckerListFullName, puckerList
from calcCoords import calcCoords, recalcCoords
from guiUtils import buttonWithIcon, HBOX_SPACING, VBOX_SPACING, BUTTON_SPACING, createRCraneWindowObject
from rotamerSeq import determineAlternateConf

DIFFERENT_PUCKER_COLOR = "pink"
BEST_PATH_COLOR = "palegreen"
BOLD_WEIGHT = pango.WEIGHT_HEAVY


class ReviewSuitesGui:
    """A class for the graphical interface used to review the predicted conformers."""
    
    def __init__(self, predictedProbs, bestPath, pseudoMol, origCoords, intermediateAtomLocs, minimizationScores, window = None, useResNums = True, oldCoordsObj = None, ignoreDensity = False, prevTorsions = None, nextTorsions = None):
        """Initialize a ReviewSuitesGui object and create a GUI for reviewing the predicted suites
        
        ARGUMENTS:
            predictedProbs       - the probabilities for each rotamer calculated for each suite
                                   formatted as a list of suites, where each suite is represented as a dictionary of rotamer => probability
            bestPath             - a list of the most likely rotamers for each suite
                                   This is typically calculed by rotamerSeq using an HMM, so it may differ slightly from simply taking all the maxima from predictedProbs
            pseudoMol            - a PseudoMolecule object for the built molecule
            origCoords           - a copy of the coordinates contained in pseudoMol as a Chain object
            intermediateAtomLocs - phosphate, O3', and C3' coordinates for each nucleotide immediately before minimization of that nucleotide was started
            minimizationScores   - the minimization scores for each nucleotide
        OPTIONAL ARGUMENTS:
            window               - a window to put the GUI into
                                   If not provided, a new window will be created.
            useResNums           - whether or not to display the rotamer numbers (as opposed to just the indices) in the GUI
                                   Defaults to True
            oldCoordsObj         - a Coot generic object for drawing the coordinates we are potentially replacing when running rotamerize
                                   If not provided, then the toggle checkbox will not be displayed and the generic object will not
                                   be destroyed when closing the window.
            ignoreDensity        - ignore the density when performing the minimization
                                   defaults to False
            prevTorsions         - the pre-rotamerize torsions for the suite before the start of origCoords.  These torsions will be used
                                   as restraints when rebuilding the first suite
                                   defaults to None
            nextTorsions         - the pre-rotamerize torsions for the suite after the end of origCoords.  These torsions will be used
                                   as restraints when rebuilding the first suite
                                   defaults to None
        RETURNS:
            an initialized ReviewSuitesGui object
        EFFECTS:
            Displays an RCrane dialog box for reviewing the predicted suites
        NOTE:
            If the pseudoMol object has savedCoordinates, then an an "Accept New Coordinates" and a "Restore Original Coordinates" will be displayed.
            Otherwise, only a "Done" button will be displayed
        """
        
        self.__predictedProbs = predictedProbs
        self.__bestPath = bestPath
        self.__pseudoMolecule = pseudoMol
        self.__origCoords = origCoords
        self.__intermediateAtomLocs = intermediateAtomLocs
        self.__window = window
        self.__useResNums = useResNums
        self.__oldCoordsObj = oldCoordsObj
        self.__minimizationScores = minimizationScores
        self.__ignoreDensity = ignoreDensity
        self.__prevTorsions = prevTorsions
        self.__nextTorsions = nextTorsions
        #print minimizationScores
        
        self.__curPath  = deepcopy(bestPath)
        self.__curSuite = 1
        self.__numNucs = len(self.__bestPath) + 2
        self.__treeViewChangedHandlerID = None
        
        #if the initial nucleotide doesn't have a C1' atom, then we can't rebuilt it
        #(This happens when we're rotamerizing.  The origCoords object contains a few 5' atoms.)
        #as a result we set __skipFirstNuc so we know to ignore the first nucleotide
        if origCoords.nucs[0].hasAtom("C1'"):
            self.__skipFirstNuc = 0
        else:
            self.__skipFirstNuc = 1
        
        #use clearCache() to initialize the cache variables
        self.__clearCache()
        
        #print origCoords
        
        #output = open("foo2.txt", 'wb')
        #dump([predictedProbs, bestPath, pseudoMol._PseudoMolecule__molecule[0][0][1], origCoords, intermediateAtomLocs, minimizationScores], output)
        #output.close()
        
        #create the review suite GUI
        self.__reviewSuites()
        
    
    def __reviewSuites(self):
        """Create the review suites GUI
        
        ARGUMENTS:
            None
        RETURNS:
            None
        EFFECTS:
            Displays an RCrane dialog box for reviewing the predicted suites
        """
        
        if self.__window is None:
            self.__window = createRCraneWindowObject()
        elif self.__window is not None:
            #remove the current contents of the window
            windowChild = self.__window.get_child()
            if windowChild is not None:
                self.__window.remove(windowChild)
            self.__window.resize(1,1) #remove any previous size constraints
        
        #connect the close window signal
        self.__window.connect("delete_event", self.__closeWin)
        
        
        # Create the buttons
        prevConfButton      = buttonWithIcon("  Previous Conformer  ",    "gtk-go-up")
        nextConfButton      = buttonWithIcon("  Next Conformer  ",        "gtk-go-down")
        
        prevSuiteButton     = buttonWithIcon("  Previous Suite  ",    "gtk-media-previous")
        nextSuiteButton     = buttonWithIcon("  Next Suite  ",        "gtk-media-next")
        
        if self.__pseudoMolecule.hasSavedCoordinates():
            doneButton     = buttonWithIcon("  Accept New Coordinates  ",      "gtk-ok")
            cancelButton   = buttonWithIcon("  Restore Original Coordinates  ", "gtk-cancel")
        else:
            doneButton     = buttonWithIcon("  Done  ",      "gtk-ok")
        
        
        #create the suiteFrame, since that needs to be passed to the previous and next suite buttons
        suiteNum = self.__origCoords.nucs[1 + self.__skipFirstNuc].resNum
        if self.__useResNums:
            frameTitle = "Suite %s (1 of %i)" % (str(suiteNum), len(self.__predictedProbs))
        else:
            frameTitle = "Suite 1 of %i" % len(self.__predictedProbs)
        
        suiteFrame = gtk.Frame(frameTitle)
        
        
        #create the TreeView objects
        (confListStore, initialConfNum) = self.__createConfListStore(1)
        (confTreeView, confTreeSelection, rowHeight) = self.__createConfTreeView(confListStore, initialConfNum)
        if len(self.__bestPath) > 1:
            overviewTreeView = self.__createOverviewTreeView(confTreeView, suiteFrame)
        else:
            overviewTreeView = None
        self.__treeViewChangedHandlerID = confTreeSelection.connect("changed", self.__changed, overviewTreeView)
        
        #create the ScrolledWindow objects used to display the TreeViews
        (treeViewScroll, treeViewScrollWidth) = self.__createConfScrolledWindow(confTreeView, rowHeight)
        if len(self.__bestPath) > 1:
            (overviewScroll, wideOverview) = self.__createOverviewScrolledWindow(overviewTreeView, prevConfButton, treeViewScrollWidth)
        
        #connect the buttons
        prevConfButton.connect("clicked", self.__prevConf, confTreeSelection)
        nextConfButton.connect("clicked", self.__nextConf, confTreeSelection)
        
        #we don't next previous and next suite buttons if there's only a single suite
        if len(self.__bestPath) > 1:
            prevSuiteButton.connect("clicked", self.__prevSuite, confTreeView, suiteFrame, overviewTreeView)
            nextSuiteButton.connect("clicked", self.__nextSuite, confTreeView, suiteFrame, overviewTreeView)
        
        doneButton.connect("clicked", self.__accept)
        
        if self.__pseudoMolecule.hasSavedCoordinates():
            cancelButton.connect("clicked", self.__restoreOldCoords)
        
        
        #put the new contents into the window
        windowBox = gtk.VBox(False, VBOX_SPACING)
        
        #create the suite overview frame if there is more than one suite
        if len(self.__bestPath) > 1:
            overviewFrame = gtk.Frame("Overview")
            overviewBox = gtk.HBox(True, 0)
            overviewBox.pack_start(overviewScroll, True, wideOverview, 0)
            overviewFrame.add(overviewBox)
            windowBox.pack_start(overviewFrame, False, True, BUTTON_SPACING)

        
        suiteFrameHBox = gtk.HBox(False, HBOX_SPACING)
        suiteFrameVBox = gtk.VBox(False, VBOX_SPACING)
        suiteFrame.add(suiteFrameVBox)
        suiteFrameVBox.pack_start(suiteFrameHBox, True, True, BUTTON_SPACING)
        windowBox.pack_start(suiteFrame, True, True, BUTTON_SPACING)
        
        prevNextConfBox = gtk.VBox(True, VBOX_SPACING)
        prevNextConfAlignment = gtk.Alignment(xalign=0.5, yalign=0.5)
        prevNextConfAlignment.add(prevNextConfBox)
        suiteFrameHBox.pack_start(prevNextConfAlignment, False, False, BUTTON_SPACING)
        prevNextConfBox.pack_start(prevConfButton, True, True, BUTTON_SPACING)
        prevNextConfBox.pack_start(nextConfButton, True, True, BUTTON_SPACING)
        
        #create a checkbox for displaying the old coordinates if we have a generic object
        if self.__oldCoordsObj is not None:
            oldCoordsCheckButton = gtk.CheckButton("Display old coordinates")
            oldCoordsCheckButton.set_active(True)
            oldCoordsCheckButton.connect("toggled", self.__toggleOldCoords)
            oldCoordsAlignment = gtk.Alignment(xalign = 0.5)
            oldCoordsAlignment.add(oldCoordsCheckButton)
            prevNextConfBox.pack_start(oldCoordsAlignment, True, True, BUTTON_SPACING)
        
        suiteFrameHBox.pack_start(treeViewScroll, True, False, BUTTON_SPACING)
        
        #we don't next previous and next suite buttons if there's only a single suite
        if len(self.__bestPath) > 1:
            prevNextSuiteBox = gtk.HBox(True, HBOX_SPACING)
            prevNextSuiteAlignment = gtk.Alignment(xalign=0.5)
            prevNextSuiteAlignment.add(prevNextSuiteBox)
            prevNextSuiteBox.pack_start(prevSuiteButton, True, True, BUTTON_SPACING)
            prevNextSuiteBox.pack_start(nextSuiteButton, True, True, BUTTON_SPACING)
            #If this is not a rotamerize call, then we put the previous and next suite buttons outside of the "Suite X of X"
            #frame.  If it is a rotamerize call, then the buttons go inside of the frame.  This is inconsistant, but it makes
            #things look a little more balanced (and it's a minor enough thing that I doubt anyone will ever notice the
            #inconsistancy)
            if self.__pseudoMolecule.hasSavedCoordinates():
                suiteFrameVBox.pack_start(prevNextSuiteAlignment, False, False, BUTTON_SPACING)
            else:
                windowBox.pack_start(prevNextSuiteAlignment, False, False, BUTTON_SPACING)
        
        #if this is a rotamerize call, then we pack the "Accept New Coordinates" and the "Restore Original Coordinates"
        #buttons
        if self.__pseudoMolecule.hasSavedCoordinates():
            doneBox = gtk.HBox(True, HBOX_SPACING)
            doneAlignment = gtk.Alignment(xalign = 0.5)
            doneAlignment.add(doneBox)
            windowBox.pack_start(doneAlignment, False, False, BUTTON_SPACING)
            doneBox.pack_start(doneButton, True, True, BUTTON_SPACING)
            doneBox.pack_start(cancelButton, True, True, BUTTON_SPACING)
            
        #if this isn't a rotamerize call, then we just pack the "Done" button
        else:
            doneAlignment = gtk.Alignment(xalign = 0.5)
            doneAlignment.add(doneButton)
            windowBox.pack_start(doneAlignment, False, False, BUTTON_SPACING)
        
        #set the current rot
        self.__selectedConf = self.__curPath[0]
        
        #go to the center of the first suite
        firstPhos = self.__pseudoMolecule.getPhosCoords(suiteNum)
        set_rotation_centre(*firstPhos)
        add_atom_label(self.__pseudoMolecule.molNum(), self.__pseudoMolecule.chain, int(suiteNum), " P  ") #label the phosphate
            #Note that add_atom_label can't handle insertion codes, so this command will label the wrong atom if we are
            #trying to center on a suite with an insertion code
        graphics_draw()
        
        #display the new window
        self.__window.add(windowBox)
        self.__window.show_all()
        
        if len(self.__bestPath) > 1:
            nextSuiteButton.grab_focus() #make the Next Suite button the default focus if it exists
        else:
            nextConfButton.grab_focus()  #otherwise, make the Next Conformer button the default focus
        
        
    def __changed(self, confTreeSelection, overviewTreeView):
        """Respond to the user selecting a different rotamer
        
        ARGUMENTS:
            confTreeSeletion - the TreeSelection object for the conformer TreeView
        RETURNS:
            None
        EFFECTS:
            Recalculates and reminimizes coordinates using the newly selected conformer for the current suite
        """
        
        #figure out what nucleotides we need to cache
        startingNucIndex = self.__curSuite + self.__skipFirstNuc - 1
        if startingNucIndex <= 0:
            startingNucIndex = 0
            hasPrevSuite = False #there's no suite before this one, so we don't have to worry about changes to the starting pucker
        else:
            hasPrevSuite = True
        endingNucIndex = self.__curSuite + self.__skipFirstNuc + 1
        #if endingNuc > self.__numNucs: endingNuc = self.__numNucs
        
        startingNucNum = self.__origCoords.nucs[startingNucIndex].resNum
        endingNucNum   = self.__origCoords.nucs[endingNucIndex].resNum
        
        #print "\n\n\n*****************__changed called\n*************************\n**************************"
        #print "startingNucIndex =", startingNucIndex
        #print "endingNucIndex =", endingNucIndex
        #print "hasPrevSuite =", hasPrevSuite        
        #print "startingNucNum =", startingNucNum
        #print "endingNucNum   =", endingNucNum
        #print "self.__prevTorsions = ", self.__prevTorsions
        #from time import sleep; sleep(3)
        
        previousConf = self.__selectedConf
        
        #determine if we need to store the previous nucleotide when cacheing the structure so that we can store its O3' position
        #(the O3' generally doesn't move much, but if a minimization fails horribly (i.e. we don't get anything chemically reasonable),
        #then the O3' will often be far out of position)
        if startingNucIndex >= 2:
            cacheStartingNucNum = self.__origCoords.nucs[startingNucIndex-1].resNum
        else:
            cacheStartingNucNum  = startingNucNum
        
        #if need be, cache the existing structure
        if not(self.__structureCache.has_key(previousConf)):
            #print "Storing structure for", previousConf
            self.__structureCache[previousConf] = self.__pseudoMolecule.getCootNucs(cacheStartingNucNum, endingNucNum)
        
        
        #figure out what the newly selected conformer is
        (confListStore, selectedConfIter) = confTreeSelection.get_selected()
        selectedConf = confListStore.get_value(selectedConfIter, 0)
        print "Conformer %s was selected" % selectedConf
        self.__selectedConf = selectedConf
        
        #see if we're changing the starting pucker (and we're not at the first suite or the first suite of a rotamerize)
        if hasPrevSuite and puckerList[previousConf][0] != puckerList[selectedConf][0] and not (startingNucIndex == 1 and self.__prevTorsions):
            
            preStartingNucNum = self.__origCoords.nucs[startingNucIndex-1].resNum
            if startingNucIndex >= 2:
                #if there's a nucleotide before the pre-starting nucleotide, then cache that too
                #since we may wind up changing the O3' atom
                preStartingNucNum = self.__origCoords.nucs[startingNucIndex-2].resNum
                #print "preStartingNucNum now includes previous previous suite"
            #print "preStartingNucNum =", startingNucNum
            
            #if we are, then see if we need to cache the current structure
            if not self.__startingPuckerStructureCache.has_key(puckerList[previousConf][0]):
                self.__startingPuckerStructureCache[puckerList[previousConf][0]] = self.__pseudoMolecule.getCootNucs(preStartingNucNum, startingNucNum)
            
            #then see if we have the structure we need in the cache
            if self.__startingPuckerStructureCache.has_key(puckerList[selectedConf][0]):
                self.__pseudoMolecule.setCootNucs(preStartingNucNum, startingNucNum, self.__startingPuckerStructureCache[puckerList[selectedConf][0]])
                changingStartingPucker = False
            else:
                changingStartingPucker = puckerList[selectedConf][0]
            
            #see if we've calculated the appropriate conformer for the previous suite
            if len(self.__previousSuiteConf) == 2:
                previousSuiteConf = self.__previousSuiteConf[puckerList[selectedConf][0]]
            else:
                #if not, then calculate it now
                self.__previousSuiteConf[puckerList[previousConf][0]] = self.__curPath[self.__curSuite-2]
                
                previousSuiteLeadingPucker = puckerList[self.__curPath[self.__curSuite-2]][0]
                previousSuiteConf = determineAlternateConf(previousSuiteLeadingPucker, puckerList[selectedConf][0], self.__curSuite-2, self.__predictedProbs)
                self.__previousSuiteConf[puckerList[selectedConf][0]] = previousSuiteConf
                
        else:
            changingStartingPucker = False
            if self.__curSuite > 1:
                previousSuiteConf = self.__curPath[self.__curSuite-2]
            elif self.__prevTorsions:
                previousSuiteConf = self.__prevTorsions
            else:
                previousSuiteConf = None
        
        #print "changingStartingPucker =", changingStartingPucker
        #print "previousSuiteConf =", previousSuiteConf
        #print "self.__previousSuiteConf = ", self.__previousSuiteConf
        
        
        #see if we're changing the ending pucker
        #print "endingNuc =", endingNuc
        #print "self.__numNucs =", self.__numNucs
        if endingNucIndex == (self.__numNucs - 1 + self.__skipFirstNuc):
            #if this is the last suite, then we don't really care what the ending pucker is when dealing with caching the next nucleotide
            #so we're only going to store one cached structure for both 2' and 3' ending pucker
            #to see if we've already stored a structure, we check to see if __endingPuckerStructureCache is a dictionary (since it gets reset to an empty dictionary in __nextSuite)
            #we also only need to cache a single nucleotide, since there's no endingNuc+1 nucleotide
            if isinstance(self.__endingPuckerStructureCache, dict):
                self.__endingPuckerStructureCache = self.__pseudoMolecule.getCootNucs(endingNucNum, endingNucNum)
            else:
                #if __endingPuckerStructureCache isn't a dictionary, then we've already stored a structure there and we should restore it
                self.__pseudoMolecule.setCootNucs(endingNucNum, endingNucNum, self.__endingPuckerStructureCache)
            
            changingEndingPucker = False
            
            if self.__nextTorsions:
                nextSuiteConf = self.__nextTorsions
            else:
                nextSuiteConf = None
            
        elif puckerList[previousConf][1] != puckerList[selectedConf][1]:
            
            postEndingNucNum = self.__origCoords.nucs[endingNucIndex+1].resNum
            #print "postEndingNucNum =", postEndingNucNum
            
            #if we're changing the ending pucker and this isn't the last suite, then see if we need to cache the current structure
            if not self.__endingPuckerStructureCache.has_key(puckerList[previousConf][1]):
                self.__endingPuckerStructureCache[puckerList[previousConf][1]] = self.__pseudoMolecule.getCootNucs(endingNucNum, postEndingNucNum)
            
            #then see if we have the structure we need in the cache
            if self.__endingPuckerStructureCache.has_key(puckerList[selectedConf][1]):
                self.__pseudoMolecule.setCootNucs(endingNucNum, postEndingNucNum, self.__endingPuckerStructureCache[puckerList[selectedConf][1]])
                changingEndingPucker = False
            else:
                changingEndingPucker = puckerList[selectedConf][1]
            
            #see if we've calculated the appropriate conformer for the next suite
            if len(self.__nextSuiteConf) == 2:
                nextSuiteConf = self.__nextSuiteConf[puckerList[selectedConf][1]]
            else:
                #if not, then calculate it now
                self.__nextSuiteConf[puckerList[previousConf][1]] = self.__curPath[self.__curSuite]
                
                nextSuiteEndingPucker = puckerList[self.__curPath[self.__curSuite]][1]
                nextSuiteConf = determineAlternateConf(puckerList[selectedConf][1], nextSuiteEndingPucker, self.__curSuite, self.__predictedProbs)
                self.__nextSuiteConf[puckerList[selectedConf][1]] = nextSuiteConf
        else:
            
            postEndingNucNum = self.__origCoords.nucs[endingNucIndex+1].resNum
            #print "postEndingNucNum =", postEndingNucNum
            
            #even if we're not changing the ending pucker, we should still reset the structure of the next nucleotide back to how it was before we started changing rotamers
            if self.__endingPuckerStructureCache.has_key(puckerList[previousConf][1]):
                self.__pseudoMolecule.setCootNucs(endingNucNum, postEndingNucNum, self.__endingPuckerStructureCache[puckerList[previousConf][1]])
            else:
                #if the structure hasn't been cached, then store the current structure.  That means that this is the first new conformer we've selected, so we the structure is already how we want it
                self.__endingPuckerStructureCache[puckerList[previousConf][1]] = self.__pseudoMolecule.getCootNucs(endingNucNum, postEndingNucNum)
            
            changingEndingPucker = False
            if self.__curSuite < (self.__numNucs - 2):
                nextSuiteConf = self.__curPath[self.__curSuite]
            elif self.__nextTorsions:
                nextSuiteConf = self.__nextTorsions
            else:
                nextSuiteConf = None
        
        #print ""
        #print "changingEndingPucker =", changingEndingPucker
        #print "nextSuiteConf =", nextSuiteConf
        #print "self.__nextSuiteConf = ", self.__nextSuiteConf
        #pprint(self.__previousSuiteConf)
        
        
        #see if we have the structure for this conformer in the cache
        if self.__structureCache.has_key(selectedConf):
            #if we do, then just use that structure
            #print "Restoring structure for", selectedConf
            #print "startingNucNum =", startingNucNum
            #print "cacheStartingNucNum =", cacheStartingNucNum            
            #print "endingNucNum   =", endingNucNum
        
            self.__pseudoMolecule.setCootNucs(cacheStartingNucNum, endingNucNum, self.__structureCache[selectedConf])
            self.__pseudoMolecule.drawExtraBonds()
        else:
            #if we don't, then we have to calculate one
            
            rotList = [previousSuiteConf, selectedConf, nextSuiteConf]
            #note that startingNuc and endingNuc start numbering at 0 (whereas Coot starts at 1, so we have to add one when we send these variables to recalcCoords)
            minimizationStartingNucIndex = self.__curSuite + self.__skipFirstNuc - 1
            minimizationEndingNucIndex = self.__curSuite + self.__skipFirstNuc 
            
            #print "minimizationStartingNucIndex =   ", minimizationStartingNucIndex
            #print "minimizationEndingNucIndex =     ", minimizationEndingNucIndex
            #print "changingStartingPucker =         ", changingStartingPucker
            #print "changingEndingPucker =           ", changingEndingPucker
            
            #see if we need to change any suites other than the current one
            if changingStartingPucker:
                rotList.insert(0, self.__curPath[self.__curSuite-3])
                minimizationStartingNucIndex -= 1
            
            if changingEndingPucker:
                if endingNucIndex < (self.__numNucs - 2 + self.__skipFirstNuc):
                    rotList.append(self.__curPath[self.__curSuite+1])
                else:
                    rotList.append(None)
                minimizationEndingNucIndex += 1
                
            minimizationStartingNucNum = self.__origCoords.nucs[minimizationStartingNucIndex].resNum
            minimizationEndingNucNum   = self.__origCoords.nucs[minimizationEndingNucIndex].resNum
            
            #reset the necessary nucleotides so that we can restart the minimization
            self.__pseudoMolecule.resetNucs(minimizationStartingNucNum, minimizationEndingNucNum, self.__origCoords, self.__intermediateAtomLocs)
            self.__pseudoMolecule.clearExtraBonds()
            
            recalcCoords(minimizationStartingNucNum, minimizationEndingNucNum, rotList, self.__origCoords, self.__pseudoMolecule, self.__window, self.__ignoreDensity)
            self.__pseudoMolecule.drawExtraBonds()
            
            #if necessary, restore the B-factors for the rebuilt atoms
            if self.__pseudoMolecule.hasSavedBfacs():
                self.__pseudoMolecule.restoreSavedBfacs(minimizationStartingNucNum, minimizationEndingNucNum)
                
        
        #update the conformers on the suite overview table
        if overviewTreeView is not None:
            self.__changeOverviewConformer(overviewTreeView, self.__curSuite, selectedConf)
            if isinstance(previousSuiteConf, basestring):
                #update the overview table if we've changed the previous conformer
                #previousSuiteConf can also be a list of torsions if we're changing the first nt of a rotamerization
                #but that shouldn't be reflected in the overview table
                self.__changeOverviewConformer(overviewTreeView, self.__curSuite-1, previousSuiteConf)
            if isinstance(nextSuiteConf, basestring):
                #update the overview table if we've changed the next conformer (ignoring lists as above)
                self.__changeOverviewConformer(overviewTreeView, self.__curSuite+1, nextSuiteConf)
    
    
    def __prevConf(self, widget, confTreeSelection):
        """Respond to the user clicking on the previous conformer button
        
        ARGUMENTS:
            widget           - the button used to invoke this function
            confTreeSeletion - the TreeSelection object for the conformer TreeView
        RETURNS:
            None
        EFFECTS:
            Highlights the previous conformer (which will implicitly invoke __changed)
        """
        
        (confListStore, selectedConfIter) = confTreeSelection.get_selected()
        
        #there is no iter_prev, so we have to jump through some hoops to get the previous row
        curPos = confListStore.get_path(selectedConfIter)[0]
        if curPos != 0:
            newIter = confListStore.get_iter((curPos-1))
            confTreeSelection.select_iter(newIter)
            
            confTreeView = confTreeSelection.get_tree_view()
            confTreeView.scroll_to_cell((curPos-1))
    
    
    def __nextConf(self, widget, confTreeSelection):
        """Respond to the user clicking on the next conformer button
        
        ARGUMENTS:
            widget           - the button used to invoke this function
            confTreeSeletion - the TreeSelection object for the conformer TreeView
        RETURNS:
            None
        EFFECTS:
            Highlights the next conformer (which will implicitly invoke __changed)
        """
        
        (confListStore, selectedConfIter) = confTreeSelection.get_selected()
        newIter = confListStore.iter_next(selectedConfIter)
        if newIter is not None:
            confTreeSelection.select_iter(newIter)
            
            confTreeView = confTreeSelection.get_tree_view()
            confTreeView.scroll_to_cell(confListStore.get_path(newIter))
    
    
    def __createConfListStore(self, curSuite):
        """Creates a ListStore containing data about the specified suite
        
        ARGUMENTS:
            curSuite - the suite to create a ListStore for
        RETURNS:
            confListStore  - a ListStore containing data about curSuite
            initialConfNum - the most likely suite (i.e. the row of confListStore to initially highlight)
        """
        
        curSuite -= 1 #the arrays start at 0, not 1
        
        confListStore = gtk.ListStore(str, str, str, str, 'gboolean', 'gboolean', 'gboolean', int)
        
        conformerList = self.__predictedProbs[curSuite].items()
        conformerList.sort(key=lambda(x): x[1], reverse=True)
        (defaultLeadingPucker, defaultEndingPucker) = puckerList[self.__bestPath[curSuite]]
        i = 0
        for (conf, score) in conformerList:
            confListStore.append([conf, "%4.2f" % score] + puckerListFullName[conf] + [(puckerList[conf][0] != defaultLeadingPucker), (puckerList[conf][1] != defaultEndingPucker), (conf == self.__bestPath[curSuite]), i+1])
            if conf == self.__curPath[curSuite]: initialConfNum = i
            i += 1
        
        return (confListStore, initialConfNum)
    
    
    def __createConfTreeView(self, confListStore, initialConfNum):
        """Create the TreeView object for the list of conformers
        
        ARGUMENTS:
            confListStore     - the ListStore object to attach to the TreeView
            initialConfNum    - the most likely suite (i.e. the row of confListStore to initially highlight)
        RETURNS:
            confTreeView      - the newly created TreeView object
            confTreeSelection - the TreeSelection object associated with the TreeView
            rowHeight         - the height of a single row in the TreeView (used to determine how high to size the ScrolledWindow)
        """
        
        #then create the columns and pack them into the tree view object
        confTreeView  = gtk.TreeView(confListStore)
        confTreeView.set_enable_search(False)
        
        renderRank = gtk.CellRendererText()
        renderRank.set_property('xalign', 0.5)
        renderRank.set_property('cell-background-gdk', self.__window.get_style().bg[gtk.STATE_NORMAL])
        columnRank = gtk.TreeViewColumn("Rank", renderRank, text=7)
        confTreeView.append_column(columnRank)
        
        renderConf = gtk.CellRendererText()
        renderConf.set_property('xalign', 0.5)
        renderConf.set_property('cell-background', BEST_PATH_COLOR)
        columnConf = gtk.TreeViewColumn("Conformer", renderConf, text=0, cell_background_set=6)
        confTreeView.append_column(columnConf)
        
        renderScore = gtk.CellRendererText()
        renderScore.set_property('xalign', 0.5)
        renderScore.set_property('cell-background', BEST_PATH_COLOR)
        columnScore = gtk.TreeViewColumn("Score", renderScore, text=1, cell_background_set=6)
        confTreeView.append_column(columnScore)
        
        renderLeadingPucker = gtk.CellRendererText()
        renderLeadingPucker.set_property('xalign', 0.5)
        columnLeadingPucker = gtk.TreeViewColumn("Leading Pucker", renderLeadingPucker, text=2)
        columnLeadingPucker.set_cell_data_func(renderLeadingPucker, self.__setPuckerBg, (4, 6))
        confTreeView.append_column(columnLeadingPucker)
        
        renderEndingPucker = gtk.CellRendererText()
        renderEndingPucker.set_property('xalign', 0.5)
        columnEndingPucker = gtk.TreeViewColumn("Ending Pucker", renderEndingPucker, text=3)
        columnEndingPucker.set_cell_data_func(renderEndingPucker, self.__setPuckerBg, (5, 6))
        confTreeView.append_column(columnEndingPucker)
        
        #select the appropriate row of the TreeView
        confTreeSelection = confTreeView.get_selection()
        confTreeSelection.select_iter(confListStore.get_iter((initialConfNum)))
        
        #figure out the height of a row (including the line dividing two rows)
        rowHeight = columnLeadingPucker.cell_get_size()[4] + confTreeView.style_get_property("vertical-separator")
        
        return (confTreeView, confTreeSelection, rowHeight)
    
    
    def __createConfScrolledWindow(self, confTreeView, rowHeight):
        """Create a ScrolledWindow object to display the conformer TreeView
        ARGUMENTS:
            confTreeView        - the TreeView object used to display the list of conformers
            rowHeight           - the height of a row in confTreeView
        RETURNS:
            treeViewScroll      - the newly created ScrolledWindow object
            treeViewScrollWidth - the width of treeViewScroll (used to figure out how wide to make the overview ScrolledWindow)
        """
        
        treeViewScroll = gtk.ScrolledWindow()
        treeViewScroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
            #we make sure that the window is big enough that it should never need a horizontal scrollbar, but sometimes GTK
            #insists on adding one, so we disallow it here
        treeViewScroll.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        treeViewScroll.add(confTreeView)
        
        #figure out the size of the treeview + vertical scrollbar is so we can set the scrolled window to the appropriate size
        (tableWidthReq, tableHeightReq) = confTreeView.size_request()
        try:
            vScrollWidthReq = treeViewScroll.get_vscrollbar().size_request()[0]
        except AttributeError:
            #get_vscrollbar only exists in PyGTK 2.8 or above
            #if we're using an older version of PyGTK, then just check the default size of a vertical scrollbar
            vScrollWidthReq = gtk.VScrollbar().size_request()[0]
        scrollbarSpacing = treeViewScroll.style_get_property("scrollbar-spacing")
        #rowHeight = columnLeadingPucker.cell_get_size()[4] + confTreeView.style_get_property("vertical-separator")
        headerHeight = tableHeightReq - len(puckerList)*rowHeight
        treeViewScrollWidth = tableWidthReq + vScrollWidthReq + scrollbarSpacing + 4
        treeViewScroll.set_size_request(treeViewScrollWidth, headerHeight + rowHeight * 3 + 4)
            #the "+ 4"'s are for the shadow, since there doesn't seem to be a GTK property corresponding to shadow size
            
        return (treeViewScroll, treeViewScrollWidth)
    
    
    def __createOverviewTreeView(self, confTreeView, suiteFrame):
        """Create a TreeView object for the suite overview
        
        ARGUMENTS:
            confTreeView      - the TreeView object for the list of conformers
            suiteFrame        - the Frame object containing information about the current suite
        RETURNS:
            overviewTreeView  - the newly created TreeView object
        """
        
        #create the overview list store
        typeList = [str] * len(self.__curPath)
        overviewListStore = gtk.ListStore(*typeList)
        
        resNumList = self.__origCoords.resNumList()[1+self.__skipFirstNuc:-1]
        overviewListStore.append(resNumList)
        overviewListStore.append([conf.replace("&", "&amp;") for conf in self.__curPath])
            #we need to escape the ampersands since we're using Pango markup
        
        #decide how wide to make the columns
        #they should be as wide as the longest residue number or conformer name (conformer names are always 2 characters)
        columnWidthChars = max([len(resNum) for resNum in resNumList] + [2])
        
        overviewTreeView = gtk.TreeView(overviewListStore)
        overviewTreeView.set_enable_search(False)
        overviewTreeView.set_headers_visible(False)
        overviewTreeView.get_selection().set_mode(gtk.SELECTION_NONE)
        
        #set_grid_lines() only exists in Gtk 2.10 or newer
        try:
            overviewTreeView.set_grid_lines(gtk.TREE_VIEW_GRID_LINES_VERTICAL)
        except AttributeError:
            #in Gtk's older than 2.10, there doesn't seem to be a way to set the gridlines to be vertical only
            #without modifying the theme, which we probably don't want to do
            #so we just leave the gridlines out
            #(The "enable-grid-lines" property can only be set to True or False, not vertical-only)
            pass

        
        setColumn = len(self.__curPath) #this column is False for the suite number row and True for the conformer row
        #we use it to selectively apply formatting changes
        
        #create columns and renderers for each suite
        for (colNum, conf) in enumerate(self.__curPath):
            #print "(colNum, conf) = ", colNum, conf
            
            renderer = gtk.CellRendererText()
            renderer.set_property('xalign', 0.5)
            renderer.set_property('width-chars', columnWidthChars+2) #the +2 ensures that we have enough width for bolding
            #renderer.set_property('width-chars', 8) #for debugging
            
            if colNum == 0:
                renderer.set_property('weight', BOLD_WEIGHT)
            

            #figure out the ratio of the best conformer score to the second best conformer score
            (bestProb, secondBestProb) = sorted(self.__predictedProbs[colNum].values(), reverse=True)[0:2]
            try:
                ratio = bestProb / secondBestProb
            except ZeroDivisionError:
                ratio = float('inf')
            
            #retrieve the minimization score for the second nucleotide of the suite
            #for the first suite, use either the first or the second nucleotide, whichever is worse
            if colNum == 0:
                minimizationScore = max(self.__minimizationScores[0], self.__minimizationScores[1])
            else:
                minimizationScore = self.__minimizationScores[colNum+1]
            
            #set the background color based on how sure we are about the prediction or how well the minimization worked (whichever is worse)
            if ratio <= 1.2 or minimizationScore >= 30:
                renderer.set_property('cell-background', 'red')
            elif ratio <= 1.75 or minimizationScore >= 20:
                renderer.set_property('cell-background', 'orange')
            elif ratio <= 2.5 or minimizationScore >= 10:
                renderer.set_property('cell-background', 'yellow')
            #otherwise, leave the background as white
            
            
            #store the column number as the title so we can easily retrieve it later
            column = gtk.TreeViewColumn(str(colNum+1), renderer, markup=colNum)
            column.set_expand(True)
            overviewTreeView.append_column(column)
            
            #make all cells the same height 
            #otherwise the height of the second row will change when the font switches from bold to bold+italics
            if colNum == 0:
                #we want the cells to be the same height that the first column starts as, since that one starts bold
                colHeight = column.cell_get_size()[4] - confTreeView.style_get_property("vertical-separator")
            renderer.set_property("height", colHeight)
        
        overviewTreeView.connect("button-press-event", self.__overviewClick, confTreeView, suiteFrame)
        
        return overviewTreeView
    
    
    def __createOverviewScrolledWindow(self, overviewTreeView, prevConfButton, treeViewScrollWidth):
        """Create a ScrolledWindow object to display the suite overview TreeView
        
        ARGUMENTS:
            overviewTreeView    - the TreeView object used to display the suite overview
            prevConfButton      - the previous conformer button (so we can calculate its width)
            treeViewScrollWidth - the width of the TreeView used to display the conformer list
        RETURNS:
            overviewScroll      - the newly created ScrolledWindow object
            wideOverview        - whether or not the ScrolledWindow takes up the full window width
        """
        
        #If the overview table is wider than the suite frame (i.e. wider than the window is going to be), then we want to
        #limit the overview table width to the size of the suite frame and include a horizontal scrollbar.
        #To know if the overview table is wider or narrower than the suite frame , first we have to determine the width of the suite frame.
        #Since the suite frame isn't realized (i.e. drawn on the screen) yet, we can just call suite_frame.size_request(),
        #so we have to add up the size requests for all of the lowest level objects.
        
        overviewScroll = gtk.ScrolledWindow()
        overviewScroll.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        overviewScroll.add(overviewTreeView)
        
        (overviewTableWidthReq, overviewTableHeightReq) = overviewTreeView.size_request()
        
        #determine the width of the previous conformer button
        prevConfButtonHBox = prevConfButton.get_child()
        #add up the width of the icon and the text
        prevConfButtonWidth = sum([child.size_request()[0] for child in prevConfButtonHBox.get_children()])
        #add on the width of the icon and text padding (the 2* is for left and right padding for each)
        prevConfButtonWidth += (2 * prevConfButtonHBox.query_child_packing(prevConfButtonHBox.get_children()[0])[2] + 2 * prevConfButtonHBox.query_child_packing(prevConfButtonHBox.get_children()[1])[2])
        
        suiteFrameWidth = treeViewScrollWidth + prevConfButtonWidth + BUTTON_SPACING + 3*HBOX_SPACING
        
        if overviewTableWidthReq <= suiteFrameWidth:
            overviewScroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
            overviewScroll.set_size_request(overviewTableWidthReq + 4, overviewTableHeightReq + 4)
            wideOverview = False
        else:
            
            #figure out how tall the scrollbar is
            try:
                hScrollHeightReq = overviewScroll.get_hscrollbar().size_request()[1]
            except AttributeError:
                #get_hscrollbar only exists in PyGTK 2.8 or above
                #if we're using an older version of PyGTK, then just check the default size of a horizontal scrollbar
                hScrollHeightReq = gtk.HScrollbar().size_request()[1]
            scrollbarSpacing = overviewScroll.style_get_property("scrollbar-spacing")
            
            overviewScroll.set_policy(gtk.POLICY_ALWAYS, gtk.POLICY_NEVER)
            overviewScroll.set_size_request(suiteFrameWidth, hScrollHeightReq + scrollbarSpacing + overviewTableHeightReq + 4)
            wideOverview = True
            
        return (overviewScroll, wideOverview)
        
    def __setPuckerBg(self, column, renderer, confListStore, iter, columnList):
        """Appropriately set the background of a cell listing sugar pucker
        
        ARGUMENTS:
            column          - the TreeViewColumn to be changed
            renderer        - the CellRenderer for the column
            confListStore   - the ListStore containing the displayed data
            iter            - a TreeIter pointing to the specified row of the TreeView
            columnList      - a list of (the number of the column to be changed, the number of the column containing the best conformer)
        RETURNS:
            None
        EFFECTS:
            Sets the background color for the specified cell to:
                DIFFERENT_PUCKER_COLOR if it's different than the pucker for the best path
                BEST_PATH_COLOR is it's the best conformer
                None otherwise
        """
        
        (puckerColumn, bestPathColumn) = columnList
        isDifferentPucker = confListStore.get_value(iter, puckerColumn)
        if isDifferentPucker:
            renderer.set_property('cell-background', DIFFERENT_PUCKER_COLOR)
        else:
            isBestPath = confListStore.get_value(iter, bestPathColumn)
            if isBestPath:
                renderer.set_property('cell-background', BEST_PATH_COLOR)
            else:
                renderer.set_property('cell-background', None)
    
    
    def __prevSuite(self, widget, confTreeView, suiteFrame, overviewTreeView):
        """Select the previous suite
        
        ARGUMENTS:
            widget           - the button used to invoke this function.  Ignored.
            confTreeView     - the TreeView used to display data about conformers for the current suite
            suiteFrame       - the GTK frame containing the "Suite X of Y" label
            overviewTreeView - the TreeView used to display the suite overview
        RETURNS:
            None
        EFFECTS:
            Recenters the screen on the previous suite and updates the RCrane GUI to reflect that suite.
        """
        if self.__curSuite > 1:
            self.__prevOrNextSuite(confTreeView, suiteFrame, overviewTreeView, self.__curSuite, self.__curSuite-1)
            
    
    def __nextSuite(self, widget, confTreeView, suiteFrame, overviewTreeView):
        """Select the next suite
        
        ARGUMENTS:
            widget           - the button used to invoke this function.  Ignored.
            confTreeView     - the TreeView used to display data about conformers for the current suite
            suiteFrame       - the GTK frame containing the "Suite X of Y" label
            overviewTreeView - the TreeView used to display the suite overview
        RETURNS:
            None
        EFFECTS:
            Recenters the screen on the next suite and updates the RCrane GUI to reflect that suite.
        """
        if self.__curSuite < len(self.__predictedProbs):
            self.__prevOrNextSuite(confTreeView, suiteFrame, overviewTreeView, self.__curSuite, self.__curSuite+1)
    
    
    def __prevOrNextSuite(self, confTreeView, suiteFrame, overviewTreeView, oldSuiteIndex, newSuiteIndex):
        """Select the specified suite
        
        ARGUMENTS:
            confTreeView     - the TreeView used to display data about conformers for the current suite
            suiteFrame       - the GTK frame containing the "Suite X of Y" label
            overviewTreeView - the TreeView used to display the suite overview
            oldSuiteIndex    - the suite we are moving from
            newSuiteIndex    - the suite we are moving to
        RETURNS:
            None
        EFFECTS:
            Recenters the screen on newSuite and updates the RCrane GUI to reflect that suite.
        """
        
        #store the newly selected conformer
        self.__curPath[oldSuiteIndex-1] = self.__selectedConf
        
        #if we've switched puckers, store the previous or next conformers
        if len(self.__previousSuiteConf) == 2:
            #print "Previous suite was", self.__curPath[oldSuiteIndex-2]
            self.__curPath[oldSuiteIndex-2] = self.__previousSuiteConf[puckerList[self.__selectedConf][0]]
            #print "Setting previous suite to", self.__curPath[oldSuiteIndex-2]
        if len(self.__nextSuiteConf) == 2:
            #print "Next suite was", self.__curPath[oldSuiteIndex]
            self.__curPath[oldSuiteIndex  ] = self.__nextSuiteConf    [puckerList[self.__selectedConf][1]]
            #print "Setting next suite to", self.__curPath[oldSuiteIndex]
        
        #go to the center of the next suite
        oldSuiteNum = self.__origCoords.nucs[oldSuiteIndex + self.__skipFirstNuc].resNum        
        newSuiteNum = self.__origCoords.nucs[newSuiteIndex + self.__skipFirstNuc].resNum
        newPhos = self.__pseudoMolecule.getPhosCoords(newSuiteNum)
        set_rotation_centre(*newPhos)
        remove_atom_label(self.__pseudoMolecule.molNum(), self.__pseudoMolecule.chain, int(oldSuiteNum), " P  ")
        add_atom_label(self.__pseudoMolecule.molNum(), self.__pseudoMolecule.chain, int(newSuiteNum), " P  ") #label the phosphate
        graphics_draw()
        
        #update the suite label
        if self.__useResNums:
            frameTitle = "Suite %s (%i of %i)" % (str(newSuiteNum), newSuiteIndex, len(self.__predictedProbs))
        else:
            frameTitle = "Suite %i of %i" % (newSuiteIndex, len(self.__predictedProbs))
        suiteFrame.set_label(frameTitle)
        
        (newListStore, initialConfNum) = self.__createConfListStore(newSuiteIndex)
        
        confTreeSelection = confTreeView.get_selection()
        confTreeSelection.disconnect(self.__treeViewChangedHandlerID)
        
        confTreeView.set_model(newListStore)
        confTreeSelection.select_iter(newListStore.get_iter((initialConfNum)))
        confTreeView.scroll_to_cell((initialConfNum), use_align=True, row_align=1.0)
        self.__treeViewChangedHandlerID = confTreeSelection.connect("changed", self.__changed, overviewTreeView)
        
        
        #update the suite overview table
        
        #move the bolding to the new column
        overviewTreeView.get_column(oldSuiteIndex-1).get_cell_renderers()[0].set_property('weight', pango.WEIGHT_NORMAL)
        overviewTreeView.get_column(newSuiteIndex-1).get_cell_renderers()[0].set_property('weight', BOLD_WEIGHT)
        
        #scroll the treeview if necessary
        overviewTreeView.scroll_to_cell((0,), overviewTreeView.get_column(newSuiteIndex-1))
        
        #force a redraw of both rows of the overview stable so that the boldness gets updated properly
        overviewModel = overviewTreeView.get_model()
        overviewModel.row_changed((0,), overviewModel.get_iter((0,)))
        overviewModel.row_changed((1,), overviewModel.get_iter((1,)))
        
        
        #set up the cache variables
        self.__selectedConf = self.__curPath[newSuiteIndex-1]
        self.__curSuite = newSuiteIndex
        self.__clearCache()
    
    def __clearCache(self):
        """Reset all data about the current suite (appropriate for when we've just switched suites)
        
        ARGUMENTS:
            None
        RETURNS:
            None
        EFFECTS:
            Resets (or initializes) all cache variables
        """
        
        self.__structureCache = {}
        self.__startingPuckerStructureCache = {}
        self.__endingPuckerStructureCache = {}
        self.__previousSuiteConf = {}
        self.__nextSuiteConf = {}
        self.__intermediateAtomLocsStartingPuckerCache = {}
        self.__intermediateAtomLocsEndingPuckerCache = {}
        self.__intermediateAtomLocsCache = {}
        
    
    
    def __closeWin(self, window, widget):
        """Close the RCrane window
        
        ARUGMENTS:
            window - the window to close
            widget - the widget used to close the window
        NOTES:
            This function simply calls __accept with the appropriate argument list
        """
        
        #self.__close(widget)
        self.__accept(widget)
    
    def __close(self, widget = None):
        """Close the RCrane window
        
        ARUGMENTS:
            widget - the widget used to close the window
        RETURNS:
            None
        EFFECTS:
            Closes the RCrane window
        """
        
        self.__pseudoMolecule.clearExtraBonds()
        if self.__oldCoordsObj is not None:
            close_generic_object(self.__oldCoordsObj)
        
        graphics_draw()
        self.__window.destroy()
    
    
    def __accept(self, widget = None):
        """Accept the new molecule and close the RCrane window
        
        ARUGMENTS:
            widget - the widget used to close the window
        RETURNS:
            None
        NOTE:
            This function re-orders atoms in the pseudomolecule (that's the "accept" part) and then calls __close
        """
        
        #perform the final molecule cleanup (reorder atoms and, if we just did a rotamerize, fix segids)
        self.__pseudoMolecule.finalMoleculeCleanup(fixSegids = bool(self.__pseudoMolecule.hasSavedCoordinates()))
        self.__close(widget)
    
    
    def __toggleOldCoords(self, widget):
        """Toggle display of the old coordinates object
        
        ARGUMENTS:
            widget - the check button that was toggled
        RETURNS:
            None
        """
        
        set_display_generic_object(self.__oldCoordsObj, int(widget.get_active()))
    
    
    def __restoreOldCoords(self, widget):
        """Set the coordinates back to the way they were before we called rotamerize
        
        ARGUMENTS:
            widget - the cancel button that was clicked to call this function.  Ignored
        RETURNS:
            None
        """
        
        self.__pseudoMolecule.restoreSavedCoordinates()
        self.__close()
    
    
    def __overviewClick(self, overviewTreeView, event, confTreeView, suiteFrame):
        """Respond to the user clicking on the overview TreeView object
        
        ARGUMENTS:
            overviewTreeView - the TreeView object for the suite overview
            event            - the click event
            confTreeView     - the TreeView object for the list of conformers
            suiteFrame       - the Frame object containing information about the current suite
        RETURNS:
            None
        """
        
        #figure out which column was just clicked on
        clickedColumn = overviewTreeView.get_path_at_pos(int(event.x), int(event.y))[1]
        
        #The suite index for each column is stored as the column title
        #(since there seems to be no easy way to get that information from the column object itself)
        newSuite = int(clickedColumn.get_title())
        
        
        if newSuite == self.__curSuite:
            #if the user clicked on the current suite, then just re-center
            newSuiteNum = self.__origCoords.nucs[newSuite + self.__skipFirstNuc].resNum
            newPhos = self.__pseudoMolecule.getPhosCoords(newSuiteNum)
            set_rotation_centre(*newPhos)
        else:
            #if the user clicked on a new suite
            #print "Changing suites in __overviewClick"
            self.__prevOrNextSuite(confTreeView, suiteFrame, overviewTreeView, self.__curSuite, newSuite)
    
    
    def __changeOverviewConformer(self, overviewTreeView, suiteNum, newConf):
        """Change the conformer listed in the overview table for the specified suite
        
        ARGUMENTS:
            overviewTreeView - the TreeView object for the suite overview
            suiteNum         - the number of the suite to change
            newConf          - the new conformer to list in the overview table
        RETURNS:
            None
        """
        
        suiteNum -= 1 #we need the suite numbers to start at 0, not 1
        
        overviewListStore = overviewTreeView.get_model()
        if newConf == self.__bestPath[suiteNum]:
            overviewListStore.set_value(overviewListStore.get_iter((1,)), suiteNum, newConf.replace("&", "&amp;"))
        else:
            overviewListStore.set_value(overviewListStore.get_iter((1,)), suiteNum, "<i>" + newConf.replace("&", "&amp;") + "</i>")
        



import numbers
# Coot and JLigand communicate via "secret" files.
# 
# Coot send to jligand the comp-ids it wants joined in *to-jligand-secret-file-name*.
#
# JLigand sends Coot a file that contains filenames for the link and
# new dictionary/restraints.
#
global to_jligand_secret_file_name
to_jligand_secret_file_name = ".coot-to-jligand-8lcs"
global from_jligand_secret_link_file_name
from_jligand_secret_link_file_name = ".jligand-to-coot-link"

# Define an environment variable for the place where jligand.jar
# resides in a canonical distribution
# Can be everywhere on windows? (maybe not in new ccp4!?) its in CCP4_BIN
#
setup_ccp4()
jligand_home_env = os.getenv("JLIGAND_HOME")
jligand_home = jligand_home_env if jligand_home_env else "."

global java_command
java_command = "java"

jligand_name = "JLigand.jar"
jligand_jar = os.path.join(os.path.normpath(jligand_home), jligand_name)
# Not yet as the old version uses different files...
if not os.path.isfile(jligand_jar):
    # lets try CCP4_BIN and JLigand.jar
    ccp4_bin = os.getenv("CCP4_BIN")
    if not ccp4_bin:
       ccp4_bin = os.getenv("CBIN")
    if ccp4_bin:
        jligand_jar = os.path.join(os.path.normpath(ccp4_bin), jligand_name)
jligand_args = ["-jar", jligand_jar, "-coot"] # for coot-enable jligand

# jligand internal parameter
global imol_jligand_link
imol_jligand_link = False

def jligand_code_file_maybe(comp_id, port):

    def jligand_standard_amino_acid_qm(comp_id):
        aa_ls = ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN",
                 "GLU", "GLY", "HIS", "ILE", "LEU", "LYS", "MET",
                 "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"]
        return comp_id in aa_ls

    # if code was read from a cif file then write a line to port like
    # "CODE TLC FILE file.cif" (with a newline).
    #
    if not jligand_standard_amino_acid_qm(comp_id):
        cif_file = coot.cif_file_for_comp_id_py(comp_id)
        if (len(cif_file) > 0):
            port.write("CODE " + comp_id + " " + \
                       "FILE " + os.path.normpath(cif_file))
            port.write("\n")

def write_file_for_jligand(res_spec_1, resname_1, res_spec_2, resname_2):

    refmac_version = refmac.get_refmac_version()
    refmac_new_enough = (refmac_version[0] >= 5 and 
                         refmac_version[1] > 6)       # newer than 5.6
    fin = open(to_jligand_secret_file_name, 'w')
    int_time = time.time()
    #print "res_spec_1:", res_spec_1
    #print "res_spec_2:", res_spec_2
    chain_id_1 = coot_utils.res_spec_to_chain_id(res_spec_1)
    chain_id_2 = coot_utils.res_spec_to_chain_id(res_spec_2)

    fin.write("CODE " + resname_1 + " " + chain_id_1 + " " + \
              str(coot_utils.res_spec_to_res_no(res_spec_1)))
    fin.write("\n")
    if refmac_new_enough:
        jligand_code_file_maybe(resname_1, fin)
    fin.write("CODE " + resname_2 + " " + chain_id_2 + " " + \
              str(coot_utils.res_spec_to_res_no(res_spec_2)))
    fin.write("\n")
    if refmac_new_enough:
        jligand_code_file_maybe(resname_2, fin)
    fin.write("TIME " + str(int_time) + "\n")
    fin.close()

def click2res_spec(click):
    return [click[2],
            click[3],
            click[4]]

# every fraction of a second look to see if
# from_jligand_secret_link_file_name has been updated.  If so, then
# run the handle_read_from_jligand_file function.
#
def start_jligand_listener():

    global from_jligand_secret_link_file_name
    global startup_mtime
    
    # BL says: why do we use the link file and not the other one??
    startup_mtime = prodrg_import.get_file_latest_time(from_jligand_secret_link_file_name)
    
    def jligand_timeout_func():

        import operator
        global startup_mtime
    
        now_time = prodrg_import.get_file_latest_time(from_jligand_secret_link_file_name)
        if isinstance(now_time, numbers.Number):
            if not isinstance(startup_mtime, numbers.Number):
                startup_mtime = now_time
                # print "just set startup-mtime to (first)", startup_mtime
                handle_read_from_jligand_file()
            else:
                if (now_time > startup_mtime):
                    startup_mtime = now_time
                    # print "just set startup-mtime to", startup_mtime
                    handle_read_from_jligand_file()
        return True # we never expire...

    gobject.timeout_add(700, with_jligand.jligand_timeout_func)


def handle_read_from_jligand_file():

    global imol_jligand_link
    global from_jligand_secret_link_file_name
    
    def bond_length(pos_1, pos_2):
        def square(x):
            return x*x
        def sub(x, y):
            return x-y

        import math
        ret = math.sqrt(sum(map(square, list(map(sub, pos_1, pos_2)))))
        return ret

    def bond_length_from_atoms(atom_1, atom_2):
        if not isinstance(atom_1, list):
            print("   WARNING:: bond_length_from_atoms: atom_1 not a list:", atom_1)
            return
        elif not isinstance(atom_2, list):
                  print("   WARNING:: bond_length_from_atoms: atom_2 not a list:", atom_2)
                  return
        else:
            return bond_length(atom_1[2],
                               atom_2[2])

    def get_dist(atom_spec_1, atom_spec_2):
        atom_1 = coot_utils.get_atom(*([imol_jligand_link] + atom_spec_1))
        atom_2 = coot_utils.get_atom(*([imol_jligand_link] + atom_spec_2))
        if not (isinstance(atom_1, list) and
                isinstance(atom_2, list)):
            return False
        else:
            return bond_length_from_atoms(atom_1, atom_2)

    if os.path.isfile(from_jligand_secret_link_file_name):
        # this file consists of 2 lines:
        # the first is a file name that contains the cif dictionary for the cif
        # the second is a LINK link that should be added to the PDB file.
        #
        fin = open(from_jligand_secret_link_file_name, 'r')
        lines = fin.read().splitlines()  # remove the newlines
        fin.close()
        if not lines:
            print("BL WARNING:: empty file", from_jligand_secret_link_file_name)
        else:
            cif_dictionary = lines[0]
            # was it the READY marker or a cif file name?
            if (cif_dictionary == "READY"):
                print("JLigand is ready")
                # maybe I need to set something here? (that
                # from now a modification of .jligand-to-coot
                # is means that we should read it?)
            if os.path.isfile(cif_dictionary):
                coot.read_cif_dictionary(cif_dictionary)
                link_line = lines[1]
                print("Now handle this link line", link_line)
                if (len(link_line) > 72):
                    atom_name_1 = link_line[12:16]
                    atom_name_2 = link_line[42:46]
                    chain_id_1  = link_line[21:22] # 1 char, FIXME in future
                    chain_id_2  = link_line[51:52]
                    res_no_1_str = link_line[22:26]
                    res_no_2_str = link_line[52:56]
                    res_no_1 = int(res_no_1_str)
                    res_no_2 = int(res_no_2_str)
                    ins_code_1 = ""  # too lazy to look these up for now.
                    ins_code_2 = ""
                    alt_conf_1 = ""  # ditto.
                    alt_conf_2 = ""
                    link_type = link_line[72]

                    print("we parsed these: ")
                    print("       atom_name_1", atom_name_1)
                    print("       atom_name_2", atom_name_2)
                    print("        chain_id_1", chain_id_1)
                    print("        chain_id_2", chain_id_2)
                    print("      res_no_1_str", res_no_1_str)
                    print("      res_no_2_str", res_no_2_str)
                    print("         link_type", link_type)

                    # check res_no_1/2 as number? Shoudl be earlier as
                    # converted to int now... use try...
                    if (coot_utils.valid_model_molecule_qm(imol_jligand_link)):
                        res_spec_1 = [chain_id_1, res_no_1, ins_code_1]
                        res_spec_2 = [chain_id_2, res_no_2, ins_code_2]
                        atom_spec_1 = res_spec_1 + [atom_name_1, alt_conf_1]
                        atom_spec_2 = res_spec_2 + [atom_name_2, alt_conf_2]
                        dist = get_dist(atom_spec_1, atom_spec_2)
                        if not dist:
                            print("bad dist %s from %s %s" %(dist, atom_spec_1, atom_spec_2))
                        else:
                            coot.make_link(imol_jligand_link, atom_spec_1,
                                      atom_spec_2, link_type, dist)
                        
                    
                    
            
        
        
        







camera.setFovy(5)
camera.setZClipFront(100)
camera.setFogFront(-100)
camera.setFogDepthRange(200)
camera.setFovy(3)
sceneSetup.getLight(0).setTranslation(CXXCoord_float(40.,40.,40,0.))

colorScheme = ColorScheme.colorByElementScheme()
colorScheme.addRule(SolidColorRule.colorRuleForSelectionStringAndName("/*/*/*.*/*:*","WHITE"))
colorScheme.addRule(SolidColorRule.colorRuleForSelectionStringAndName('("A") & [C]',"CORAL"))
colorScheme.addRule(SolidColorRule.colorRuleForSelectionStringAndName('("U") & [C]',"RED"))
colorScheme.addRule(SolidColorRule.colorRuleForSelectionStringAndName('("G") & [C]',"CYAN"))
colorScheme.addRule(SolidColorRule.colorRuleForSelectionStringAndName('("C") & [C]',"LAWNGREEN"))

import pypdb
pdbCodes = ["1ffk"]
for pdbCode in pdbCodes:
    pdbContents = pypdb.get_pdb_file(pdbCode)
    molecule = MyMolecule.createFromString(pdbContents)
    sceneSetup.setTranslation(molecule.getCentre()*-1.)
    molinst3=MolecularRepresentationInstance.create(molecule,ColorScheme.colorChainsScheme(),"ALL","DishyBases")
    sceneSetup.addRepresentationInstance(molinst3)
    molinst4=MolecularRepresentationInstance.create(molecule,ColorScheme.colorChainsScheme(),"ALL","Ribbon")
    molinst4.getRepresentation().updateFloatParameter("ribbonStyleCoilThickness",1.)
    molinst4.getRepresentation().updateFloatParameter("ribbonStyleDNARNAWidth",2.)
    sceneSetup.addRepresentationInstance(molinst4)
    self.wfWidget.repaint()
    



#!/usr/bin/env python
"""Common constants and functions used in the various GUI modules"""

# Copyright 2010, 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os.path
import gtk

#spacing parameters for GTK boxes
VBOX_SPACING = 5
HBOX_SPACING = 10
BUTTON_SPACING = 2

rcranePath = os.path.dirname(os.path.abspath(__file__))
    
def buttonWithIcon(label, image):
    """Create a button object with the specified label and GTK stock image.
    
    ARGUMENTS:
        label  - the text label for the button
        image  - the name of the GTK stock image
    RETURNS:
        button - the constructed button object
    NOTE:
        The buttons created here are different than the stock buttons created by gtk.Button().
        Buttons created by gtk.Button() ignore the label if it does not match the stock image and
        resize differently than buttons created here.
    """
    
    button = gtk.Button()
    hbox = gtk.HBox(False, 0)
    #hbox.set_border_width(2)
    buttonImage = gtk.image_new_from_stock(image, gtk.ICON_SIZE_BUTTON)
    buttonLabel = gtk.Label(label)
    hbox.pack_start(buttonImage, False, False, 3)
    hbox.pack_start(buttonLabel, False, False, 3)
    button.add(hbox)
    
    return button


def createRCraneWindowObject(title = "RCrane"):
    """Create a new window object with the appropriate name and icon
    
    OPTIONAL ARGUMENTS:
        title - the title for the window.  Defaults to "RCrane"
    RETURNS:
        window - a window object
    """
    
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    window.set_title(title)
    window.set_icon_from_file(os.path.join(rcranePath, "icon.svg"))
    return window


def createRCraneErrorDialog(errorText):
    """Create a RCrane error dialog popup
    
    ARGUMENTS:
        errorText - the text of the error message to report to the user
    RETURNS:
        None
    """
    
    errorDialog = gtk.MessageDialog(type = gtk.MESSAGE_ERROR, buttons = gtk.BUTTONS_OK)
    errorDialog.set_title("RCrane Error")
    errorDialog.set_markup(errorText)
    errorDialog.run()
    errorDialog.destroy()



#  Copyright 2012 by the University of Oxford
#  Copyright 2014 by Medical Research Council
#  Author: Paul Emsley
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or (at
#  your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA

import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk

# import pygtk, gtk, pango
import xml.etree.ElementTree as et
import string
import math
import types
import gzip
import coot

class PDB_Entry:

    def __init__(self, entry_attrib, xml_file_name):
        self.xml_file_name = xml_file_name
        self.pdbid                 = False
        self.Fo_Fc_correlation     = False
        self.IoverSigma            = False
        self.numMillerIndices      = False
        self.percent_RSRZ_outliers = False
        self.absolute_percentile_percent_RSRZ_outliers = False
        self.relative_percentile_percent_RSRZ_outliers = False
        self.percent_RSRZ_outliers                     = False
        self.acentric_outliers     = False
        self.centric_outliers      = False
        self.EDS_R                 = False
        self.EDS_resolution        = False
        self.EDS_resolution_low    = False
        self.DataAnisotropy        = False
        self.DataCompleteness      = False
        self.Fo_Fc_correlation     = False
        self.WilsonBaniso          = False
        self.WilsonBestimate       = False
        self.relative_percentile_RNAsuiteness = False
        self.absolute_percentile_RNAsuiteness = False
        self.RNAsuiteness          = False
        self.TwinL                 = False
        self.TwinFraction          = False
        self.TwinL2                = False
        self.xtriage_input_columns = False
        self.TransNCS              = False
        self.DCC_R                 = False
        self.DCC_refinement_program = False
        self.DCC_Rfree              = False
        self.absolute_percentile_DCC_Rfree = False
        self.relative_percentile_DCC_Rfree = False
        self.CCP4version        = False
        self.RefmacVersion      = False
        self.bulk_solvent_k     = False
        self.bulk_solvent_b     = False
        self.absolute_percentile_percent_rama_outliers = False
        self.absolute_percentile_percent_rota_outliers = False
        self.relative_percentile_percent_rama_outliers = False
        self.relative_percentile_percent_rota_outliers = False
        self.percent_rama_outliers                     = False
        self.percent_rota_outliers                     = False
        self.RestypesNotcheckedForBondAngleGeometry    = False

        try:
            self.pdbid = entry_attrib['pdbid']
        except KeyError as e:
            pass

        try:
            self.absolute_percentile_percent_RSRZ_outliers = entry_attrib['absolute-percentile-percent-RSRZ-outliers']
            self.relative_percentile_percent_RSRZ_outliers = entry_attrib['relative-percentile-percent-RSRZ-outliers']
            self.percent_RSRZ_outliers                     = entry_attrib['percent-RSRZ-outliers']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.acentric_outliers                         = entry_attrib['acentric_outliers']
            self.centric_outliers                          = entry_attrib['centric_outliers']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.EDS_R              = entry_attrib['EDS_R']
            self.EDS_resolution     = entry_attrib['EDS_resolution']
            self.EDS_resolution_low = entry_attrib['EDS_resolution_low']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.DataAnisotropy                            = entry_attrib['DataAnisotropy']
            self.DataCompleteness                          = entry_attrib['DataCompleteness']
            self.Fo_Fc_correlation                         = entry_attrib['Fo_Fc_correlation']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.WilsonBaniso                              = entry_attrib['WilsonBaniso']
            self.WilsonBestimate                           = entry_attrib['WilsonBestimate']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.relative_percentile_RNAsuiteness          = entry_attrib['relative-percentile-RNAsuiteness']
            self.absolute_percentile_RNAsuiteness          = entry_attrib['absolute-percentile-RNAsuiteness']
            self.RNAsuiteness                              = entry_attrib['RNAsuiteness']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.TwinL              = entry_attrib['TwinL']
            self.TwinFraction       = entry_attrib['TwinFraction']
            self.TwinL2             = entry_attrib['TwinL2']
        except KeyError as e:
            # it is OK if a Entry doesnt have these
            pass

        try:
            self.xtriage_input_columns                     = entry_attrib['xtriage_input_columns']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass

        try:
            self.TransNCS                                  = entry_attrib['TransNCS']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass

        try:
            self.DCC_R                                     = entry_attrib['DCC_R']
            self.DCC_refinement_program                    = entry_attrib['DCC_refinement_program']
            self.DCC_Rfree                                 = entry_attrib['DCC_Rfree']
            self.absolute_percentile_DCC_Rfree             = entry_attrib['absolute-percentile-DCC_Rfree']
            self.relative_percentile_DCC_Rfree             = entry_attrib['relative-percentile-DCC_Rfree']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass

        try:
            self.CCP4version                               = entry_attrib['CCP4version']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass

        try:
            self.RefmacVersion      = entry_attrib['RefmacVersion']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass

        try:
            self.bulk_solvent_k     = entry_attrib['bulk_solvent_k']
            self.bulk_solvent_b     = entry_attrib['bulk_solvent_b']
        except KeyError as e:
            # it is OK if a Entry doesnt have one
            pass


class ModelledSubgroup:

    # <bond-outlier atom0="CD" atom1="CE" mean="1.508" obs="1.561" stdev="0.025" z="2.11"/>
    class bond_outlier:
        def __init__(self, bond_attrib):
            # print 'bond_attrib: ', bond_attrib
            self.atom0  = False
            self.atom1  = False
            self.mean   = False
            self.obs    = False
            self.stddev = False
            self.z      = False
            try:
                self.atom0  = bond_attrib['atom0']
                self.atom1  = bond_attrib['atom1']
                self.mean   = bond_attrib['mean']
                self.obs    = bond_attrib['obs']
                self.stdev  = bond_attrib['stdev']
                self.z      = bond_attrib['z']
            except KeyError as e:
                pass

    # <mog-bond-outlier Zscore="8.78" atoms="C2',N1'" mean="1.46" mindiff="0.10" numobs="26" obsval="1.32" stdev="0.02"/>
    class mog_bond_outlier:
        def __init__(self, mog_bond_attrib):
            # print 'mog_bond_attrib: ', mog_bond_attrib
            self.Zscore  = False
            self.atoms   = False
            self.mean    = False
            self.mindiff = False
            self.numobs  = False
            self.obsval  = False
            self.stdev   = False
            try:
                self.Zscore  = mog_bond_attrib['Zscore']
                self.atoms   = mog_bond_attrib['atoms']
                self.mean    = mog_bond_attrib['mean']
                self.mindiff = mog_bond_attrib['mindiff']
                self.numobs  = mog_bond_attrib['numobs']
                self.obsval  = mog_bond_attrib['obsval']
                self.stdev   = mog_bond_attrib['stdev']
            except KeyError as e:
                pass

    # <angle-outlier atom0="C4" atom1="C5" atom2="C7" mean="119.000" obs="121.304" stdev="0.600" z="3.84"/>
    class angle_outlier:
        def __init__(self, angle_attrib):
            # print 'angle_attrib: ', angle_attrib
            self.atom0  = False
            self.atom1  = False
            self.atom2  = False
            self.mean   = False
            self.obs    = False
            self.stddev = False
            self.z      = False
            try:
                self.atom0  = angle_attrib['atom0']
                self.atom1  = angle_attrib['atom1']
                self.atom2  = angle_attrib['atom2']
                self.mean   = angle_attrib['mean']
                self.obs    = angle_attrib['obs']
                self.stdev  = angle_attrib['stdev']
                self.z      = angle_attrib['z']
            except KeyError as e:
                pass

    # <mog-angle-outlier Zscore="4.09" atoms="O1',C1,C2" mean="107.14" mindiff="0.92" numobs="97"
    #                    obsval="113.79" stdev="1.63"/>
    class mog_angle_outlier:
        def __init__(self, mog_angle_attrib):
            # print 'mog_angle_attrib: ', mog_angle_attrib
            self.Zscore  = False
            self.atoms   = False
            self.mean    = False
            self.mindiff = False
            self.numobs  = False
            self.obsval  = False
            self.stdev   = False
            try:
                self.Zscore  = mog_angle_attrib['Zscore']
                self.atoms   = mog_angle_attrib['atoms']
                self.mean    = mog_angle_attrib['mean']
                self.mindiff = mog_angle_attrib['mindiff']
                self.numobs  = mog_angle_attrib['numobs']
                self.obsval  = mog_angle_attrib['obsval']
                self.stdev   = mog_angle_attrib['stdev']
            except KeyError as e:
                pass


    # <clash atom="O" cid="116" clashmag="0.52" dist="2.09"/>
    class clash:
        def __init__(self, clash_attrib):
            # print 'clash_attrib: ', clash_attrib
            self.atom     = False
            self.cid      = False
            self.clashmag = False
            self.dist     = False
            try:
                self.atom     = clash_attrib['atom']
                self.cid      = clash_attrib['cid']
                self.clashmag = clash_attrib['clashmag']
                self.dist     = clash_attrib['dist']
            except KeyError as e:
                pass


    def __init__(self, subgroup):

        # individual bits that may or may not be there:
        self.bond_outliers      = []
        self.angle_outliers     = []
        self.mog_bond_outliers  = []
        self.mog_angle_outliers = []
        self.clashes            = []
        self.rsrz     = False
        self.RNAscore = False
        self.suite    = False
        self.flippable_sidechain = None # i.e. unset ATM.
        self.rama = False
        self.num_H_reduce = False

        subgroup_attrib = subgroup.attrib
        for child in subgroup:
            if child.tag == 'bond-outlier':
                bo = self.bond_outlier(child.attrib)
                if bo:
                    self.bond_outliers.append(bo)
            if child.tag == 'mog-bond-outlier':
                bo = self.mog_bond_outlier(child.attrib)
                if bo:
                    self.mog_bond_outliers.append(bo)
            if child.tag == 'angle-outlier':
                ao = self.angle_outlier(child.attrib)
                if ao:
                    self.angle_outliers.append(ao)
            if child.tag == 'mog-angle-outlier':
                ao = self.mog_angle_outlier(child.attrib)
                if ao:
                    self.mog_angle_outliers.append(ao)
            if child.tag == 'clash':
                c = self.clash(child.attrib)
                if c:
                    self.clashes.append(c)

        try:
            self.altcode   = subgroup_attrib['altcode']
            self.resnum    = subgroup_attrib['resnum']
            self.chain     = subgroup_attrib['chain']
            self.icode     = subgroup_attrib['icode']
            self.resname   = subgroup_attrib['resname']
            self.model     = subgroup_attrib['model']
            self.rscc      = False
            self.NatomsEDS = False
            self.owab      = False
            self.avgoccu   = False
            self.rsr       = False
            self.rama      = False

            try:
                self.rsr   = subgroup_attrib['rsr']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a rsrz
                pass

            try:
                self.avgoccu   = subgroup_attrib['avgoccu']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a rsrz
                pass

            try:
                self.rama      = subgroup_attrib['rama']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a rama
                pass

            try:
                self.owab = subgroup_attrib['owab']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a owab
                pass

            try:
                self.NatomsEDS = subgroup_attrib['NatomsEDS']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a NatomsEDS
                pass

            try:
                self.rscc = subgroup_attrib['rscc']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a rscc
                pass

            try:
                self.rsrz = subgroup_attrib['rsrz']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a rsrz
                pass
            try:
                self.RNAscore = subgroup_attrib['RNAscore']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a RNAscore
                pass
            try:
                self.suite = subgroup_attrib['suite']
            except KeyError as e:
                # it is OK if a ModelledSubgroup doesnt have a suite
                pass

            return None

        except KeyError as e:
            print('ModelledSubgroup: KeyError with key', e)

    # Is this a residue for which we want to create a button?
    #
    # return a string (for the button, if yes) otherwise return False.
    #
    def is_problematic_p(self):

       is_bad = False
       problems = []
       for bo in self.bond_outliers:
           try:
               if abs(float(bo.z)) > 2:
                   is_bad = True
                   s = "Bond Outlier " + bo.atom0 + " " + bo.atom1 + " z = " + bo.z
                   problems.append(s)
           except TypeError as e:
               pass

       for ao in self.angle_outliers:
           try:
               if abs(float(ao.z)) > 3:
                   is_bad = True
                   s = "Angle Outlier " + ao.atom0 + " " + ao.atom1 + " " + ao.atom2 + " z = " + ao.z
                   problems.append(s)
           except TypeError as e:
               pass

       for bo in self.mog_bond_outliers:
           try:
               if abs(float(bo.Zscore)) > 2:
                   is_bad = True
                   s = "Mogul-based Bond Outlier " + bo.atoms + ", z = " + bo.Zscore
                   problems.append(s)
           except TypeError as e:
               pass

       for ao in self.mog_angle_outliers:
           try:
               if abs(float(ao.Zscore)) > 2:
                   is_bad = True
                   s = "Mogul-based Angle Outlier " + ao.atoms + ", z = " + ao.Zscore
                   problems.append(s)
           except TypeError as e:
               pass

       for clash in self.clashes:
           try:
               if (clash.clashmag):
                   if float(clash.clashmag) >= 0.4:
                       is_bad = True
                       s = "Clash atom " + clash.atom + " score: " + clash.clashmag
                       problems.append(s)
           except TypeError as e:
               pass

       if self.rsrz:
           if float(self.rsrz) > 0.40:
               is_bad = True
               s = "Bad RSRZ " + self.rsrz
               problems.append(s)

       if self.RNAscore:
           if float(self.RNAscore) < 0.5:
               is_bad = True
               s = "Bad RNAScore " + self.RNAscore
               problems.append(s)

       if self.rama:
           if self.rama == 'OUTLIER':
               is_bad = True
               s = "Ramachandran Outlier"
               problems.append(s)

       if is_bad:
           return problems
       else:
           return False

# return validation info, which wraps entry_validation_info
#
def parse_wwpdb_validation_xml(xml_string):

    try:
        tree = et.fromstring(xml_string)
        modelled_subgroups = []
        for child in tree:
            if child.tag == 'Entry':
                xml_file_name = 'Fake-XML-Filename'
                entry_info = PDB_Entry(child.attrib, xml_file_name)
            if child.tag == 'ModelledSubgroup':
                subgroup = ModelledSubgroup(child)
                if subgroup:
                    modelled_subgroups.append(subgroup)
        validation_info = [entry_info, modelled_subgroups]
        return validation_info
    except IOError as e:
        print("error:", e)
        return False


class validation_entry_to_canvas:

    def __init__(self, entry_validation_info_in):

        self.entry_validation_info = entry_validation_info_in;
        self.bar_length = 300
        self.bar_height = 10
        # self.x_bar_offset = 130
        self.x_bar_offset = 160
        self.y_bar_offset =  20
        self.y_pixels_bar_spacing = 30
        self.setup_colour_bar_buff()
        self.abs_bar_width = 6
        self.abs_bar_height = int(self.bar_height * 1.6)
        self.vbox = False

        if self.entry_validation_info != False:
            window = Gtk.Window()
            title = "PDB Validation Report for " # ...
            if self.entry_validation_info.pdbid:
                 title += self.entry_validation_info.pdbid
            else:
                 title += self.entry_validation_info.xml_file_name
            window.set_title(title)
            self.vbox = Gtk.VBox(False, 0)
            self.vbox.set_border_width(5)
            h_sep = Gtk.HSeparator()
            da = Gtk.DrawingArea()
            da.set_size_request(560,280)
            close_button = Gtk.Button("  Close  ")
            hbox = Gtk.HBox(False, 0)
            self.vbox.append(da)
            self.vbox.append(h_sep)
            hbox.append(close_button)
            self.vbox.pack_end(hbox, False, False, 6)
            window.add(self.vbox)
            window.show_all()
            # da.connect("expose-event", self.on_drawing_area_expose)
            da.connect("configure-event", self.on_drawing_area_expose)
            close_button.connect("clicked", lambda a : window.destroy())
            self.pangolayout = da.create_pango_layout("")

    def setup_colour_bar_buff(self):

        c = 3*self.bar_length*self.bar_height*['\0']
        for j in range(self.bar_length):

            f_j = float(j)/float(self.bar_length)
            # we need g to go 255:255:0
            r = int(255*(1-math.pow(f_j, 5)))
            # we need g to go 0:255:0
            g_1 = f_j                #  0 : 0.5 : 1
            g_2 = 2 * (g_1 - 0.5)    # -1 : 0.  : 1
            g_3 = g_2 * g_2          #  1 : 0.  : 1
            g_4 = 1 - g_3            #  0 : 1.  : 0
            g = int(240*g_4)
            b = int(255*math.pow(f_j, 0.2))
            for i in range(self.bar_height):
                idx = 3*(self.bar_length*i + j)
                c[idx  ] = chr(r)
                c[idx+1] = chr(g)
                c[idx+2] = chr(b)
        # self.colour_bar_buff = string.join(c, '') 20211021-PE 
        # is this what I want (now)?
        # print("debug:: c", c)
        self.colour_bar_buff = ""
        self.colour_bar_buff.join(c)
        print("debug:: self.colour_bar_buff", self.colour_bar_buff)

    def on_drawing_area_expose(self, da, event):

        style = da.get_style()
        print(dir(style))
        # there is no fg_gc in style
        gc = style.fg_gc[Gtk.STATE_NORMAL]

        n_sliders = self.draw_sliders(da, gc)
        self.draw_top_labels(da, gc)
        self.draw_bottom_labels(da, gc, n_sliders) # Worse, Better
        self.draw_key(da, gc, n_sliders) # percentile box descriptions
        self.draw_eds_resolution_text(da, gc, n_sliders)


    def draw_eds_resolution_text(self, da, gc, n_sliders):
        if (self.entry_validation_info.EDS_resolution):
            pangolayout = da.create_pango_layout("")
            font_desc = pango.FontDescription('Sans 8')
            pangolayout.set_font_description(font_desc)
            s = 'EDS Resolution: ' + self.entry_validation_info.EDS_resolution
            pangolayout.set_text(s)
            y_level  = self.y_bar_offset + self.y_pixels_bar_spacing * n_sliders + 90
            da.window.draw_layout(gc, 10, y_level, pangolayout)


    def draw_rgb_image(self, da, gc, x, y):

        da.window.draw_rgb_image(gc, x, y, self.bar_length, self.bar_height,
                                 Gtk.gdk.RGB_DITHER_NONE, self.colour_bar_buff, self.bar_length*3)

    def bar_for_abs(self, abs_percent, y_min, da, gc):

        x = int(self.x_bar_offset + 0.01 * abs_percent * self.bar_length - 0.5 * self.abs_bar_width)
        y = int(y_min - self.bar_height * 0.25)
        da.window.draw_rectangle(gc, True, x, y, self.abs_bar_width, self.abs_bar_height)

    def arcs_for_rel(self, rel_percent, y_min, da, gc):
        x = int(self.x_bar_offset + 0.01 * rel_percent * self.bar_length - 0.5 * self.abs_bar_width)
        y = int(y_min - self.bar_height * 0.25)
        da.window.draw_rectangle(gc, False, x, y, self.abs_bar_width, self.abs_bar_height)

    # Worse, Better
    def draw_bottom_labels(self, da, gc, n_sliders):
        x_worse  = self.x_bar_offset
        y_worse  = self.y_bar_offset + self.y_pixels_bar_spacing * n_sliders + 15
        x_better = self.x_bar_offset + self.bar_length - 32
        y_better = y_worse

        pl_wb = da.create_pango_layout("")
        pl_wb.set_text('Worse')
        font_desc = pango.FontDescription('Sans 8')
        font_desc.set_style(pango.STYLE_ITALIC)
        pl_wb.set_font_description(font_desc)
        da.window.draw_layout(gc, x_worse,  y_worse,  pl_wb)

        pl_wb.set_text('Better')
        da.window.draw_layout(gc, x_better, y_better, pl_wb)

    # percentile box descriptions
    def draw_key(self, da, gc, n_sliders):
        x_key_box_abs =  self.x_bar_offset
        y_key_box_abs = self.y_bar_offset + self.y_pixels_bar_spacing * (n_sliders + 1) + 10

        x_key_box_rel = x_key_box_abs
        y_key_box_rel = y_key_box_abs + 20

        x_key_1 =  x_key_box_abs + 10
        y_key_1  = y_key_box_abs

        x_key_2 =  x_key_1
        y_key_2  = y_key_1 + 20

        da.window.draw_rectangle(gc, True, x_key_box_abs, y_key_box_abs,
                                 self.abs_bar_width, self.abs_bar_height)
        da.window.draw_rectangle(gc, False, x_key_box_rel, y_key_box_rel,
                                 self.abs_bar_width, self.abs_bar_height)

        pl = da.create_pango_layout("")
        font_desc = pango.FontDescription('Sans 9')
        pl.set_font_description(font_desc)

        pl.set_text('Percentile relative to all x-ray structures')
        da.window.draw_layout(gc, x_key_1, y_key_1, pl)
        pl.set_text('Percentile relative to x-ray structures of similar resolution')
        da.window.draw_layout(gc, x_key_2, y_key_2, pl)


    # return True if the bar for the absolute percentile was drawn,
    # otherwise return False
    #
    def draw_slider(self, name, x_for_rj, abs_str, rel_str, value_str, slider_no, da, gc):

        y = self.y_bar_offset + self.y_pixels_bar_spacing*(slider_no+1)

        # colour bar
        self.draw_rgb_image(da, gc, self.x_bar_offset, y)

        local_gc = gc;
        local_gc.set_foreground(local_gc.get_colormap().alloc_color("#888888"))

        da.window.draw_rectangle(local_gc, False, self.x_bar_offset, y, self.bar_length, self.bar_height)
        local_gc.set_foreground(local_gc.get_colormap().alloc_color("#000000"))

        # Metric text
        pangolayout = da.create_pango_layout("")
        pangolayout.set_justify(1)
        pangolayout.set_alignment(pango.ALIGN_RIGHT)
        pangolayout.set_text(name)
        pangolayout.context_changed()
        da.window.draw_layout(gc, 4+x_for_rj, y-6, pangolayout)

        # Values text
        if isinstance(value_str, types.StringType):
            x_for_value = self.x_bar_offset + self.bar_length + 12
            pangolayout.set_text(value_str)
            # print "value", x_for_value, y
            da.window.draw_layout(gc, x_for_value, y-6, pangolayout)

        # Bars for percentile scores
        try:
            if rel_str != False:
                rel = float(rel_str)
                self.arcs_for_rel(rel, y, da, gc)
            if abs_str != False:
                abs = float(abs_str)
                self.bar_for_abs( abs, y, da, gc)
                return True

            rel = float(rel_str)
        except TypeError as e:
            print(e)

        # hopefully we don't get here.
        return False


    # pass abs values
    def draw_connecting_lines(self, pc_ranks, da, gc):

        if len(pc_ranks) > 1:
            for slider_no in range(len(pc_ranks)-1):
                y_min_1 = 20 + 30*(slider_no+1)
                y_min_2 = 20 + 30*(slider_no+2)
                abs_1 = float(pc_ranks[slider_no])
                abs_2 = float(pc_ranks[slider_no+1])

                x_1 = int(self.x_bar_offset + 0.01 * abs_1 * self.bar_length)
                y_1 = int(y_min_1 + self.bar_height + 1)

                x_2 = int(self.x_bar_offset + 0.01 * abs_2 * self.bar_length)
                y_2 = int(y_min_2 - self.bar_height * 0.25)

                da.window.draw_line(gc, x_1, y_1, x_2, y_2)


    def draw_sliders(self, da, gc):

        save_abs = []
        icount_slider = 0

        abs = self.entry_validation_info.absolute_percentile_DCC_Rfree
        rel = self.entry_validation_info.relative_percentile_DCC_Rfree
        value = self.entry_validation_info.DCC_Rfree
        if abs != False and rel != False:
            state = self.draw_slider("Rfree", 114, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        abs   = self.entry_validation_info.absolute_percentile_clashscore
        rel   = self.entry_validation_info.relative_percentile_clashscore
        value = self.entry_validation_info.clashscore
        if abs != False and rel != False:
            state = self.draw_slider("Clashscore", 78, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        abs   = self.entry_validation_info.absolute_percentile_percent_rama_outliers
        rel   = self.entry_validation_info.relative_percentile_percent_rama_outliers
        value = self.entry_validation_info.percent_rama_outliers
        if abs != False and rel != False:
            state = self.draw_slider("Ramachandran Outliers", 0, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        abs = self.entry_validation_info.absolute_percentile_percent_rota_outliers
        rel = self.entry_validation_info.relative_percentile_percent_rota_outliers
        value = self.entry_validation_info.percent_rota_outliers
        if abs != False and rel != False:
            state = self.draw_slider("Sidechain Outliers", 32, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        abs = self.entry_validation_info.relative_percentile_percent_RSRZ_outliers
        rel = self.entry_validation_info.absolute_percentile_percent_RSRZ_outliers
        value = self.entry_validation_info.percent_RSRZ_outliers
        if abs != False and rel != False:
            state = self.draw_slider("RSRZ Outliers", 56, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        abs = self.entry_validation_info.relative_percentile_RNAsuiteness
        rel = self.entry_validation_info.absolute_percentile_RNAsuiteness
        value = self.entry_validation_info.RNAsuiteness
        if abs != False and rel != False:
            state = self.draw_slider("RNASuiteness", 57, abs, rel, value, icount_slider, da, gc)
            if state:
                save_abs.append(abs)
            icount_slider += 1

        self.draw_connecting_lines(save_abs, da, gc)
        return icount_slider

    def draw_top_labels(self, da, gc):
        pangolayout = da.create_pango_layout("")

        pangolayout = da.create_pango_layout("")
        font_desc = pango.FontDescription('Sans 13')
        pangolayout.set_font_description(font_desc)

        pangolayout.set_text('Metric')
        y_level = 15
        da.window.draw_layout(gc, 100, y_level, pangolayout)
        pangolayout.set_text('Percentile Ranks')
        da.window.draw_layout(gc, 245, y_level, pangolayout)
        pangolayout.set_text('Value')
        da.window.draw_layout(gc, 470, y_level, pangolayout)

def add_residue_buttons(subgroups, vbox, imol):

    def go_to_residue(button, residue_spec):
        # print "go to imol ", imol, "residue", residue_spec
        set_go_to_atom_molecule(imol)
        set_go_to_atom_from_res_spec(residue_spec)

    if vbox:
        vbox_residue_buttons = Gtk.VBox(False, 0)
        scrolled_win = Gtk.ScrolledWindow()
        # scrolled_win.set_policy(Gtk.POLICY_AUTOMATIC, Gtk.POLICY_ALWAYS)
        scrolled_win.add_with_viewport(vbox_residue_buttons)
        scrolled_win.set_size_request(-1, 200)
        for group in subgroups:
            p = group.is_problematic_p()
            if (p):
                ri_string = 'Residue '
                ri_string += group.chain
                ri_string += ' '
                ri_string += group.resnum
                ri_string += ' '
                ri_string += group.resname
                ri_string += ':'
                is_first = True
                for p_i in p:
                    ri_string += '\n'
                    ri_string += '    '
                    ri_string += p_i
                residue_button = Gtk.Button(ri_string)
                if group.icode == ' ':
                    group.icode = ''
                try:
                    r_n = int(group.resnum)
                    residue_spec = [group.chain, r_n, group.icode ]
                    residue_button.connect("clicked", go_to_residue, residue_spec)
                    vbox_residue_buttons.append(residue_button)
                    residue_button.show()
                except ValueError:
                    print('problem parsing', group.chain, group.resnum, group.icode)
        vbox.append(scrolled_win)
        scrolled_win.show()
        vbox_residue_buttons.show()

def validation_to_gui(entry_validation_info, subgroups, imol):
    vi = validation_entry_to_canvas(entry_validation_info)
    add_residue_buttons(subgroups, vi.vbox, imol)

def sort_subgroups(subgroups):
    return subgroups

def pdb_validate(accession_code, imol):

    try:
        if len(accession_code) == 4:
            middle_letters = accession_code[1] + accession_code[2]
            gz_file_name = "pdb-validation-" + accession_code + ".xml.gz"
            url = 'http://ftp.ebi.ac.uk/pub/databases/pdb/validation_reports/'
            url += middle_letters
            url += '/'
            url += accession_code
            url += '/'
            url += accession_code
            url += '_validation.xml.gz'
            status = coot.coot_get_url(url, gz_file_name)
            # turn the gz_file_name into a string
            vi = False
            try:
                gz = gzip.open(gz_file_name)
                xml_string = gz.read()
                vi = parse_wwpdb_validation_xml(xml_string)
                gz.close()
            except IOError as e:
                print("Error", e)
            if vi:
                entry_validation_info = vi[0]
                subgroups = vi[1]
                ss = sort_subgroups(subgroups)
                validation_to_gui(entry_validation_info, ss, imol)
            else:
                print("something went wrong when getting", url)
                s = "Something went wrong"
                add_status_bar_text(s)

        else:
            print('WARNING:: invalid accession code', accession_code)
    except KeyError as e:
        print("Failure to get validation for ", accession_code)

# xml_list = ["3NPQ.xml", "1FV2.xml", "2PE5.xml", "1HAK.xml", "2FGG.xml", "1CBS.xml"]

# xml_list = ["1FV2.xml"]
# xml_list = ["2PE5.xml"]
# xml_list = ["1HAK.xml"]
# xml_list = ["1CBS.xml"]
# xml_list = ["1CBS-1.xml"]
# xml_list = ["2FGG.xml"]
# # xml_list = ["3NPQ.xml", "1FV2.xml", "2PE5.xml", "1HAK.xml", "2FGG.xml", "1CBS.xml"]
# # xml_list = ["3NPQ.xml"] # DNA only?
# # xml_list = ["3NPQ.xml", "1FV2.xml", "2PE5.xml", "1HAK.xml", "2FGG.xml", "1CBS.xml"]

# for xml_file in xml_list:
#     vi = parse_wwpdb_validataion_xml(xml_file)
#     if vi:
#         entry_validation_info = vi[0]
#         subgroups = vi[1]
#         ss = sort_subgroups(subgroups)
#         validation_to_gui(entry_validation_info, ss)

# # coot_real_exit(0)
# gtk.main()

# pdb_validate('1cbs')





import numbers


# Is this true? Dont understand this
def get_stable_release_from_server_string(stri):
    return stri

# needs testing!
def get_stable_release_from_coot_version():
    s = coot.coot_version()
    ls = s.split(" ")
    return ls[0]

# return True or False (None on error, maybe should return some Error!?)
# although if no return, then var becomes None anyway.. FIXME).
def new_version_on_server(stri, is_pre_release):

    from types import StringType
    if is_pre_release:

        # pre-releases are handle with svn revision numbers (as ints).
        #
        server_rev = coot_utils.get_revision_from_string(stri)
        if server_rev:
            return server_rev > coot.svn_revision()
        else:
            return None

    else:
        # For a stable release: 
        #
        # The Coot release number (e.g. "0.6.4") is a string that can
        # be compared lexographically.
        #
        server_version = get_stable_release_from_server_string(stri)
        this_build_version = get_stable_release_from_coot_version()
        if (server_version is StringType):
            return server_version > this_build_version
    return None

# show the dialog
# 
def notify_of_new_version(stri, use_curl=False):
    #print "notify_of_new_version given string:", stri

    # Maybe protect from being called again (download is running)
    download_binary_dialog(coot_utils.coot_split_version_string(stri), use_curl)

# version_string is something like: "coot-0.6-pre-1-revision-2060"
def download_binary_dialog(version_string, use_curl=False):

    import os
    import operator
    global pending_install_in_place
    global stop_download
    global file_name_for_progress_bar
    stop_download = False
    file_name_for_progress_bar = False
    if (pending_install_in_place ==  "cancelled"):
        pending_install_in_place = False
    
    #print "BL DEBUG:: in download binary with version", version_string
    #print "BL DEBUG:: download_binary_dialog with", version_string, use_curl

    def do_download_dialog(use_curl=False):
        #
        # print "running download-binary-dialog with version-string arg:", version_string
        #
        def update_progress_bar(progress_bar):
            if file_name_for_progress_bar:
                curl_info = curl_progress_info(file_name_for_progress_bar)
                if curl_info:
                    v1 = curl_info['content-length-download']
                    v2 = curl_info['size-download']                        
                    if isinstance(v1, numbers.Number):
                        if isinstance(v2, numbers.Number):
                            f = v2 / v1
                            #print "count %s, active_count %s, f: %s" %(count, active_count, f)
                            progress_bar.set_fraction(f)

        def set_progress_bar_full(progress_bar):
            progress_bar.set_fraction(1)

        def set_file_name_func(file_name):
            global file_name_for_progress_bar
            file_name_for_progress_bar = file_name

        def pending_install_in_place_func(val):
            global pending_install_in_place
            pending_install_in_place = val

        def delete_event(*args):
            # only close window (BL says:: maybe should be hide and a global window
            # in case we want to go back to it!?)
            window.destroy()
            return False

        def cancel_event(*args):
            # stop the download process(es) if
            # they were running (they get set to
            # False when they are not running)
            # 
            global pending_install_in_place
            if file_name_for_progress_bar:
                coot.stop_curl_download(file_name_for_progress_bar)
            pending_install_in_place = "cancelled"  # not fail!
            return False

        def ok_button_event(*args):
            global pending_install_in_place
            import operator

            # only start when ok button is pressed
            gobject.timeout_add(500, idle_func) # update every 500ms is god enough?!
            if not isinstance(revision, numbers.Number):
                coot.info_dialog("Failed to communicate with server")
            else:
                # BL says:: check if we have a download available already?! (from before)
                if (version_string in get_latest_pending_version()):
                    # installer already here (win only?!)
                    pending_install_in_place = True
                else:
                    # download
                    ok_button.set_sensitive(False)
                    #align.show()
                    progress_bar.show()
                    cancel_button.show()
                    def threaded_func():
                        global pending_install_in_place
                        ret = run_download_binary_curl(revision, version_string,
                                                       pending_install_in_place_func,
                                                       set_file_name_func,
                                                       progress_bar,
                                                       use_curl)

                        if ((not ret) and
                            (not pending_install_in_place == "cancelled")):
                            print("run_download_binary_curl failed")
                            pending_install_in_place = "fail"

                    coot_gui.run_python_thread(threaded_func, [])

        #  and a timeout checking the progress of the download:
        def idle_func():
            global pending_install_in_place
            if (pending_install_in_place == "fail"):
                window.destroy()
                coot.info_dialog("Failure to download and install binary")
                return False  # stop idle func
            if (pending_install_in_place == "cancelled"):
                window.destroy()
                return False  # stop idle func when download cancelled
            if (pending_install_in_place == "full"):  # dont go yet!? Why?
                set_progress_bar_full(progress_bar)
            if pending_install_in_place:
                window.destroy()
                restart_dialog()
                return False  # stop idle func
            else:
                update_progress_bar(progress_bar)
                return True   # continue running

        s = "   New revision available " + \
            "for this binary type:   \n" + \
            coot_sys_build_type()     + \
            "\n"                      + \
            "\n"                      + \
            version_string
        revision = coot_utils.get_revision_from_string(version_string)

        window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        dialog_name = "Download "
        if coot_utils.is_windows():
            dialog_name += "installer"
        else:
            dialog_name += "binary"
        main_vbox = gtk.VBox(False, 6)
        cancel_button = gtk.Button("  Cancel  ")
        close_button = gtk.Button("  Close  ")
        ok_button = gtk.Button("  Download and Pre-install ")
        buttons_hbox = gtk.HBox(False, 6)
        progress_bar = gtk.ProgressBar()
        h_sep = gtk.HSeparator()
        info_string = gtk.Label(s)

        # needed?
        #align = gtk.Alignment(0.5, 0.5, 0, 0)
        #align.add(pbar)

        window.set_title(dialog_name)
        buttons_hbox.pack_start(ok_button, True, False, 6)
        buttons_hbox.pack_start(cancel_button, True, False, 6)
        buttons_hbox.pack_start(close_button, True, False, 6)

        main_vbox.pack_start(info_string,  True, False, 6)
        main_vbox.pack_start(h_sep,        True, False, 6)
        main_vbox.pack_start(buttons_hbox, True, False, 6)
        main_vbox.pack_start(progress_bar, True, False, 6)
        #main_vbox.pack_start(align,        True, False, 6)
        main_vbox.set_border_width(6)

        window.add(main_vbox)
        window.set_border_width(6)

        cancel_button.connect("clicked", cancel_event)
        close_button.connect("clicked", delete_event)
        ok_button.connect("clicked", ok_button_event)
        window.show_all()
        progress_bar.hide()
        #align.hide()
        cancel_button.hide()
        idle_func()  # call once for to get existing updates

    # main line
    coot_prefix = os.getenv("COOT_PREFIX")
    coot_prefix = os.path.normpath(coot_prefix) # needed for WinCoot?! FIXME
    if not coot_utils.directory_is_modifiable_qm(coot_prefix):
        if not coot_prefix:
            coot.info_dialog("COOT_PREFIX is not set.  Download not started.")
        else:
            coot.info_dialog("Directory " + coot_prefix + \
                        " is not modifiable.\n" + \
                        "Download/install not started.")
    else:
        do_download_dialog(use_curl)
        if (pending_install_in_place == "cancelled"):
            pending_install_in_place = False  # reset for next run
        
def restart_dialog(extra_text=""):

    def delete_event(*args):
        window.destroy()
        return False

    def ok_button_event(*args):
        coot_command = "coot"
        coot_args =[]
        if coot_utils.is_windows():
            # we need the name of the installer, should be in pending-install
            coot_command = get_latest_pending_version()  # shall be full path name
            prefix_dir = os.getenv("COOT_PREFIX")

            coot_args =["/instdir=" + prefix_dir,
                        "/startdir=" + os.getcwd(),
                        "/autoupdate"]
            
        # create a coot background subprocess
        coot_utils.run_concurrently(coot_command, coot_args)
            
        window.destroy()
        coot.coot_real_exit(0)

    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    dialog_name = "Restart Required to complete installation"
    main_vbox = gtk.VBox(False, 6)
    cancel_button = gtk.Button("  Later  ")
    ok_button = gtk.Button("  Restart Now  ")
    label = gtk.Label("  Restart required to complete install " + extra_text)
    buttons_hbox = gtk.HBox(False, 6)
    h_sep = gtk.HSeparator()

    window.set_title(dialog_name)
    window.add(main_vbox)
    window.set_border_width(6)
    main_vbox.pack_start(label,        False, False, 6)
    main_vbox.pack_start(h_sep,        False, False, 5)
    main_vbox.pack_start(buttons_hbox, False, False, 6)
    buttons_hbox.pack_start(ok_button, False, False, 6)
    buttons_hbox.pack_start(cancel_button, False, False, 6)

    cancel_button.connect("clicked", delete_event)
    ok_button.connect("clicked", ok_button_event)

    window.show_all()


# Paul's test binary?
# http://www.biop.ox.ac.uk/coot/software/binaries/pre-releases/coot-0.6-pre-1-revision-2535-binary-Linux-i386-centos-4-gtk2.tar.gz


def check_for_updates_gui(use_curl=False):

    from types import StringType
    import time
    global server_info_status

    server_info_status = False

    def handle_latest_version_server_response(txt_from_server):
        global server_info_status
        # OK, so the server said something.
        # Set the status here, so that the
        # function that looks to see whether
        # or not the server responded is
        # notified.
        #
        if ("The requested URL" and "was not found on this server" in txt_from_server):
            server_info_status = "file-not-found"
        else:
            server_info_status = txt_from_server
    
    def get_server_info_status_thread():
        url = coot_utils.make_latest_version_url()
        print("INFO:: get URL", url)
        if use_curl:
            # non pythonic
            #x=get_url_as_string(url) # FIXME to trick the firewall
            latest_version_server_response = coot_get_url_as_string(url)
        else:
            # pythonic version
            import urllib.request, urllib.parse, urllib.error
            try:
                latest_version_server_response = urllib.request.urlopen(url).read()
            except:
                coot.info_dialog("Could not establish connection to get latest version on server")
                return
        try:
            handle_latest_version_server_response(latest_version_server_response)
        except:
            print("BL INFO:: problem getting server response from for url", url)            # for now we give file-not-found, there should be some other form
            # of error
            handle_latest_version_server_response("The requested URL was not found on this server")

    # main line
    #
    coot_gui.run_python_thread(get_server_info_status_thread, ())
    is_pre_release = coot_utils.pre_release_qm()  # BL: why?
    global count
    count = 0
    def timeout_count():
        global count
        global server_info_status
        if (count > 2000):  # try for 20 seconds, otherwise timeout.
            # fail_with_timeout
            print("final fail: server_info_status:", server_info_status)
            # maybe some info here too?!!? 
            return False  # stop running this idle function
        elif (server_info_status == "file-not-found"):
            s = "No " + \
                ("pre-release" if is_pre_release else "release") + \
                " binary for this system (" + \
                coot_sys_build_type() + \
                ") on the binary server."
            coot.info_dialog(s)
            return False  # stop running idle function
        elif (server_info_status == ""):
            # BL says:: supposedly - havent checked
            # this happens when coot can't get past the proxy
            # 
            coot.info_dialog("Can't communicate with server")
        elif server_info_status:
            if (new_version_on_server(server_info_status, is_pre_release)):
                notify_of_new_version(server_info_status, use_curl)
            else:
                s = "No version newer than this revision (" + \
                    str(coot.svn_revision())                     + \
                    ")."
                coot.info_dialog(s)
            return False # stop running idle function
        else:
            time.sleep(0.010)
            # print "server_info_status", server_info_status
            count += 1
            return True
            
    #run_with_gtk_threading(timeout_count)
    gobject.idle_add(timeout_count)

# returns latest installer version in pending-install or empty string ""
# if no installer for windows use only!
def get_latest_pending_version():
    if coot_utils.is_windows():
        import glob
        prefix_dir = os.getenv("COOT_PREFIX")
        pending_dir = os.path.join(prefix_dir, "pending-install")
        if os.path.isdir(pending_dir):
            installer_glob = os.path.join(pending_dir, "WinCoot*.exe")
            installer_filenames = glob.glob(installer_glob)
            installer_name = ""
            if installer_filenames:
                installer_name = installer_filenames[0]
            if (len(installer_filenames) > 1):
                # shall use the newest?!
                tmp_time = 0
                for filename in installer_filenames:
                    mtime = os.path.getmtime(filename)
                    if mtime > tmp_time:
                        tmp_time = mtime
                        installer_name = filename
                # now delete the other ones
                for filename in installer_filenames:
                    if (filename != installer_name):
                        os.remove(filename)
            return installer_name
    return ""

# hack to check if a pending install for WinCoot?!
# if so, show the restart dialog?!
# not sure yet how to do it otherwise (without chaning runwincoot.bat)
if coot_utils.is_windows():
    pending_version_full = get_latest_pending_version()
    if pending_version_full:
        pending_version = os.path.basename(pending_version_full)
        restart_dialog("\n\nFound %s to install\n" %pending_version + \
                       "Please \"Restart Now\"")
    



#!/usr/bin/env python
"""A class for storing information about individual nucleotides."""

# Copyright 2010, 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from strucCalc import torsion, dist, distToLine

SUGARATOM = "C1'"

class Nucleotide:
    """A class for storing information about individual nucleotides (which may be part of a Chain object)."""
    
    def __init__(self, type, atoms, resNum = None):
        """Initialize a Nucleotide object
        
        ARGUMENTS:
            type  - the nucleotide type (i.e. A, C, G, or U)
            atoms - a hash of atom coordinates (in the form atomName: [x, y, z])
        OPTIONAL ARGUMENTS:
            resNum - the residue number (possibly including an insertion code)
        RETURNS:
            an initialized Nucleotide object
        """
        
        self.type = type
        self.atoms = atoms
        self.resNum = resNum
        self.chain = None
        self.seqNum = None
    
    
    def prevNuc(self):
        """If this nucleotide is part of a chain, return the previous nucleotide in the chain
        
        ARGUMENTS:
            None
        RETURNS:
            The Nucleotide object of the previous nucleotide if there is one
            None if this is the first nucleotide of the chain,or if this nucleotide is not part of a chain.
        """
        
        if self.seqNum > 0:
            return self.chain.nucleotides[self.seqNum-1]
        else:
            return None
    
    
    def nextNuc(self):
        """If this nucleotide is part of a chain, return the next nucleotide in the chain
        
        ARGUMENTS:
            None
        RETURNS:
            The Nucleotide object of the next nucleotide if there is one
            None if this is the last nucleotide of the chain or if this nucleotide is not part of a chain.
        """
        
        if self.seqNum < self.chain.len():
            return self.chain.nucleotides[self.seqNum+1]
        else:
            return None
    
    
    def connectedToPrev(self):
        """Is this nucleotide connected to the previous nucleotide of the chain?
        
        ARGUMENTS:
            None
        RETURNS:
            True if this nucleotide is connected to the previous nucleotide of the chain
            False if it isn't or if this nucleotide is not part of a chain.
        NOTE:
            Because the chain module does not currently support unconnected nucleotides, this function only
            tells you if this is the first nucleotide of the chain or not.
        """
        
        return (self.seqNum > 0)
    
    
    def connectedToNext(self):
        """Is this nucleotide connected to the next nucleotide of the chain?
        
        ARGUMENTS:
            None
        RETURNS:
            True if this nucleotide is connected to the previous nucleotide of the chain
            False if it isn't
            Note that the behavior of this function (unlike that of connectedToPrev()) is not defined
                if this nucleotide is not part of a chain.  (Currently, it raises an error.)
        NOTE:
            Because the chain module does not currently support unconnected nucleotides, this function only
            tells you if this is the last nucleotide of the chain or not.
        """
        
        return (self.seqNum < (self.chain.len()-1))
    
    
    def hasAtom(self, atomName):
        """Does this nucleotide contain the specified atom?
        
        ARGUMENTS:
            atomName - the name of the atom to check for
        RETURNS:
            True if this nucleotide contains atomName, False otherwise
        """
        
        return self.atoms.has_key(atomName)
    
    def eta(self):
        """Calculate the eta pseudotorsion for the current nucleotide
        
        ARGUMENTS:
            None
        RETURNS:
            The value of eta if it is defined, None otherwise.
        NOTE:
            Calculating eta requires the sugar atom of the previous nucleotide and the phosphate atom
            of the next nucleotide.  As a result, behaviour of this function is undefined if this nucleotide
            is not part of a chain.  (Currently, it raises an error.)
        """
        
        if (self.connectedToPrev() and self.connectedToNext()
              and self.prevNuc().hasAtom(SUGARATOM)
              and self.hasAtom("P")
              and self.hasAtom(SUGARATOM)
              and self.nextNuc().hasAtom("P")):
            
            return torsion(self.prevNuc().atoms[SUGARATOM],
                               self.atoms["P"],
                               self.atoms[SUGARATOM],
                               self.nextNuc().atoms["P"])
        else:
            return None
    
    
    def theta(self):
        """Calculate the theta pseudotorsion for the current nucleotide
        
        ARGUMENTS:
            None
        RETURNS:
            The value of theta if it is defined, None otherwise.
        NOTE:
            Calculating theta requires the phosphate and sugar atom of the next nucleotide.
            As a result, behaviour of this function is undefined if this nucleotide
            is not part of a chain.  (Currently, it raises an error.)
        """
        
        if (self.connectedToNext()
              and self.hasAtom("P")
              and self.hasAtom(SUGARATOM)
              and self.nextNuc().hasAtom("P")
              and self.nextNuc().hasAtom(SUGARATOM)):
            
            return torsion(self.atoms["P"],
                               self.atoms[SUGARATOM],
                               self.nextNuc().atoms["P"],
                               self.nextNuc().atoms[SUGARATOM])
        else:
            return None
    
    
    def pperp(self):
        """Calculate the base-phosphate perpendicular distance (P-perp distance) for the current nucleotide
        
        ARGUMENTS:
            None
        RETURNS:
            The P-perp distance if it is defined, None otherwise
        NOTE:
            Calculating P-perp requires the phosphate of the next nucleotide.  As a result, behaviour of this
            function is undefined if this nucleotide is not part of a chain.  (Currently, it raises an error.)
        """
        
        
        #determine the coordinates of the base nitrogen
        if self.hasAtom("N9"):
            baseN = self.atoms["N9"]
        elif self.hasAtom("N1"):
            baseN = self.atoms["N1"]
        else:
            baseN = None
        
        if baseN is not None and self.hasAtom(SUGARATOM) and self.connectedToNext() and self.nextNuc().hasAtom("P"):
            return distToLine(self.nextNuc().atoms["P"],
                              self.atoms[SUGARATOM],
                              baseN)
        else:
            return None
    
    
    def phosDist(self):
        """Calculate the phosphate-phosphate distance
        
        ARGUMENTS:
            None
        RETURNS:
            The distance from the phosphate of this nucleotide to the phosphate of the next nucleotide
            None if no next phosphate if present
        NOTE:
            Behavior of this function is undefined if this nucleotide is not part of a chain.
            (Currently, it raises an error.)
        """
        
        if self.connectedToNext() and self.hasAtom("P") and self.nextNuc().hasAtom("P"):
            return dist(self.atoms["P"], self.nextNuc().atoms["P"])
        else:
            return None
    
    
    def startingSugarDist(self):
        """Calculate the sugar-sugar distance for the starting suite of this nucleotide
        
        ARGUMENTS:
            None
        RETURNS:
            the distance from the sugar atom (i.e. C1') of the previous nucleotide to the sugar atom (C1') of this nucleotide
            None if there is no previous nucleotide
        NOTE:
            Behavior of this function is undefined if this nucleotide is not part of a chain.
            (Currently, it returns None.)
        """
        
        if self.connectedToPrev() and self.hasAtom(SUGARATOM) and self.prevNuc().hasAtom(SUGARATOM):
            return dist(self.atoms[SUGARATOM], self.prevNuc().atoms[SUGARATOM])
        else:
            return None
    
    
    def endingSugarDist(self):
        """Calculate the sugar-sugar distance for the ending suite of this nucleotide
        
        ARGUMENTS:
            None
        RETURNS:
            the distance from the sugar atom (i.e. C1') of this nucleotide to the sugar atom (C1') of the next nucleotide
            None if there is no previous nucleotide
        NOTE:
            Behavior of this function is undefined if this nucleotide is not part of a chain.
            (Currently, it raises an error.)
        """
        
        if self.connectedToNext():
            return self.nextNuc().startingSugarDist()
        else:
            return None
    
    def __str__(self):
        """Create a string representation of the object"""
        
        retVal  = ("Nuc Num:  " + str(self.resNum) + "\n")
        retVal += ("Seq Num:  " + str(self.seqNum) + "\n")        
        retVal += ("Nuc type: " + str(self.type) + "\n")        
        retVal += "Atoms: \n"
        for (name, coords) in self.atoms.iteritems():
            retVal += ("\t" + name + ":  " + str(coords) + "\n")
        return retVal
    
    
    def alpha(self):
        """Calculate the alpha torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of alpha if it is defined, None otherwise.
        """
        
        if (self.connectedToPrev()
              and self.prevNuc().hasAtom("O3'")
              and self.hasAtom("P")
              and self.hasAtom("O5'")
              and self.hasAtom("C5'")):
            
            return torsion(self.atoms["C5'"],
                           self.atoms["O5'"],
                           self.atoms["P"],
                           self.prevNuc().atoms["O3'"])
                
        else:
            return None

    def beta(self):
        """Calculate the beta torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of beta if it is defined, None otherwise.
        """
        
        if   (self.hasAtom("P")
          and self.hasAtom("O5'")
          and self.hasAtom("C5'")
          and self.hasAtom("C4'")):
            
            return torsion(self.atoms["P"],
                           self.atoms["O5'"],
                           self.atoms["C5'"],
                           self.atoms["C4'"])
                
        else:
            return None
        
    def gamma(self):
        """Calculate the gamma torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of gamma if it is defined, None otherwise.
        """
        
        if   (self.hasAtom("O5'")
          and self.hasAtom("C5'")
          and self.hasAtom("C4'")
          and self.hasAtom("C3'")):
            
            return torsion(self.atoms["O5'"],
                           self.atoms["C5'"],
                           self.atoms["C4'"],
                           self.atoms["C3'"])
                
        else:
            return None
        
    def delta(self):
        """Calculate the delta torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of delta if it is defined, None otherwise.
        """
        
        if   (self.hasAtom("C5'")
          and self.hasAtom("C4'")
          and self.hasAtom("C3'")
          and self.hasAtom("O3'")):
            
            return torsion(self.atoms["C5'"],
                           self.atoms["C4'"],
                           self.atoms["C3'"],
                           self.atoms["O3'"])
                
        else:
            return None
    
    def epsilon(self):
        """Calculate the epsilon torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of epsilon if it is defined, None otherwise.
        """
        
        if (self.connectedToNext()
              and self.hasAtom("C4'")
              and self.hasAtom("C3'")
              and self.hasAtom("O3'")
              and self.nextNuc().hasAtom("P")):
            
            return torsion(self.atoms["C4'"],
                           self.atoms["C3'"],
                           self.atoms["O3'"],
                           self.nextNuc().atoms["P"])
                
        else:
            return None
        
    
    def zeta(self):
        """Calculate the zeta torsion
        
        ARGUMENTS:
            None
        RETURNS:
            The value of zeta if it is defined, None otherwise.
        """
        
        if (self.connectedToNext()
              and self.hasAtom("C3'")
              and self.hasAtom("O3'")
              and self.nextNuc().hasAtom("P")
              and self.nextNuc().hasAtom("O5'")):
            
            return torsion(self.atoms["C3'"],
                           self.atoms["O3'"],
                           self.nextNuc().atoms["P"],
                           self.nextNuc().atoms["O5'"])
                
        else:
            return None
    
